

/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/ 
 
/* CHANGES:
 *     15Aug88 ISI: Order GenOpPtr entries wrong
 *     17Nov88 ISI: Add eqf (alias EQ) object
 *     19Nov88 ISI: Add field reference count, ComputeCuminv().
 *     21Nov88 ISI: On attach with same obj name, verify parm list.
 *     22Nov88 ISI: Attach appends rather than prepends to field chains
 *     22Nov88 ISI: Add SetEtherType()
 *     28Nov88 ISI: Generalize pseudo-program for case construction
 *     11Mar89 Merit: Added omitted initialization statement to Get_SOBJ()
 *     25Oct89 ISI: Change names of some routines
 *       Rel 2.4:
 *     01Nov89 ISI: Add new -byte objects
 *     29Nov89 ISI: Fix init bug in Get_SOBJ().
 */
 
/*
 *                          attach.c
 *
 *     This file contains the routines used for performing an attach
 *     command in statspy.  Specifically, the main routines included
 *     here are:
 * 
 *            Init_attach()
 *            Attach_SOBJ()
 *            Install_it()
 *            doIf()
 *            doAppend()
 *            doNull()
 *            CleanUp()
 *
 */
 
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "stat.h" 
#include "sobj.h"

extern struct FI_info Fields[];  /* Everything there is to know about fields*/
extern SOBJ *SOBJ_list ; /* Head of list of SOBJ's */
extern char *savestring() ;
extern char *SOBJ_error ;
extern char  Error_msg[256];
extern int   Xdebug;

     /* Transfer Vectors */
extern GENERICOP FOop, HIop, P2op, FAop, BPop, FA2op, RFop, SFop, MAop, MSop,
    MA2op, WSop, EQop, FABop, FOBop, MABop, MSBop;   

GENERICOP *GenOpPtr[] = {   
             /*** MUST MATCH class_name/class_nick in scan.c ****/    
    &FOop, &HIop, &P2op, &FAop, &BPop, &FA2op, &RFop, &SFop, &MAop, &WSop,
    &MSop, &MA2op, &EQop, &FABop, &FOBop, &MABop, &MSBop}; 
        
SOPC *inv_stack[100];   /* Stack for building invocation structure */
int   inv_stkptr = 0;   /* Stack pointer for invocation structure */
#define PUSH  inv_stack[inv_stkptr++]   
#define POP   inv_stack[--inv_stkptr] 
#define TOS   inv_stack[inv_stkptr-1]
#define TLIST if_sopcp->sopc_sublist[WRRET_TRUE]
#define FLIST if_sopcp->sopc_sublist[WRRET_FALSE]
        
       
    
    /***********************************************************
     * 
     *    ROUTINES TO CREATE INTERPRETIVE CODE FOR CONFIGURATION
     *
     *      doIf(), doAppend(), doNull(), Install_it() 
     *
     **********************************************************/

/*
 *   Construct structure for an IF statement (filter)
 *
 *      This looks complicated, because building an efficient and
 *      correct execution sequence, considering parse tree, IS 
 *      complicated! 
 *
 *   There are three implicit operands, all on the stack: the IF
 *   invocation, the TRUE alternative list, and the FALSE alternative
 *   list.  These are popped off the stack, and the resulting list
 *   is pushed back onto the stack.
 *
 *   If this fails, it sets SOBJ_error and returns FALSE. It also
 *   makes sure all outstanding invocations are on the stack, for
 *   CleanUp.
 */
boolean doIf(isnot) 
    int  isnot;
    {
    SOPC *sp, *tp, *Ancest_sp;
    SOPC *if_sopcp;
    SOPC **spp, *new_if, *nextp;
    int rc;
    SOPC *Get_SOPC();
            
    
        /* Mark True and False invocations and then merge them
         *   into a single sorted list.
         */
    for (sp = inv_stack[inv_stkptr - ((isnot)?1:2)]; sp; sp = sp->sopc_next)
        sp->sopc_rc = 1;
    for (sp = inv_stack[inv_stkptr - ((isnot)?2:1)]; sp; sp = sp->sopc_next)
        sp->sopc_rc = -1;
    doAppend();
    
    sp = POP;
    if_sopcp = POP;
    PUSH = NULL;  /* Push NULL head of chain of filter invocations */
    
        /* Now process sorted list of subordinate invocations... */
    while (sp) {
        nextp = sp->sopc_next;
        
        if ((rc = LEQtree(sp, if_sopcp)) > 0) {
            /* Subordinate == ancestor of filter invocation. Invoke at
             *    filter field, put subordinate in TRUE/FALSE field.
             */
             new_if = if_sopcp;
        }
        else if ((rc = LEQtree(if_sopcp, sp)) > 0)
            {
            /* Filter is ancestor of subordinate.  Scan list of filter
             *  invocations already built.  If find one invoked at same
             *  field, use it.  Else, look for ancestral evaluation and
             *  go indirect from it.  Else, must build new filter evaluation.
             *     [Note the importance of starting with a sorted
             *      list... we don't want an ancestor to appear LATER]
             */
                    /* Search IF list */
            new_if = Ancest_sp = NULL;                    
            for (tp = TOS; tp; tp= tp->sopc_next) {
                if (tp->sopc_AtFIno == sp->sopc_AtFIno) {
                            /* Found matching evaluation */
                    new_if = tp;
                    break;
                }
                else if (tp->sopc_indir == NULL /* must not be indirect */ &&
                       LEQtree(tp, sp) > 0) {
                               /* Make current subordinate indirect */
                     Ancest_sp = tp;
                }
            }
            if (new_if == NULL) {  
                            /* no matching IF. Must create SOPC */
                if ((new_if = Get_SOPC( 
                            if_sopcp->sopc_FIptr1, if_sopcp->sopc_FIptr2,
                            if_sopcp->sopc_sobjp, "", 2)) == NULL) {
                   /* Push rest of list onto stack for cleanup*/
                   
                    PUSH = sp;
                    return(FALSE);
                }
                new_if->sopc_indir = Ancest_sp;
                new_if->sopc_AtFIno = sp->sopc_AtFIno; 
                PUSH = new_if;
                doAppend();
            }
        }
        else if (rc < 0) {
            /* Funny pseudo-field ?? Don't know what to do here... */
        }
        else {
                /* Subordinate is cousin of IF filter -- CONFIG ERROR!
                 */
             sprintf(SOBJ_error = Error_msg, 
                   "Config Error: impossible combination of fields: %s %s",
                 Fields[if_sopcp->sopc_AtFIno].field_name,
                 Fields[sp->sopc_AtFIno].field_name);
             PUSH = sp;  /* Push rest of list onto stack for cleanup*/ 
             return(FALSE);              
        }
        
            /* Finally, chain subordinate onto TRUE/FALSE chain of
             *   the IF invocation which we have selected/constructed.
             */
        spp = (sp->sopc_rc > 0)? &new_if->sopc_sublist[WRRET_TRUE]:
                                 &new_if->sopc_sublist[WRRET_FALSE];
        sp->sopc_next = *spp;
        *spp = sp;
            /* And go on to next subordinate in original list.*/
        sp = nextp;
    }
        /* Finally, chain or delete original IF invocation. */
    if (TLIST == NULL && FLIST == NULL) 
        DeleteSOPC(if_sopcp);
    else {
        PUSH = if_sopcp;
        doAppend();
    }   
    return(TRUE);
}
    

/* 
 *  Merge two sorted lists from top of stack into one list on top of stack.
 *    All three are ordered by attach field number.
 */
boolean doAppend()
    {
    register SOPC *L1 = POP, *L2 = POP;
    SOPC *newL = NULL;
    register SOPC *nxtL, **Lpp = &newL;
    
    while (L1||L2) {
        if ((L1)&&(L2 == NULL|| L1->sopc_AtFIno <= L2->sopc_AtFIno)) {
            nxtL = *Lpp = L1;
            L1 = L1->sopc_next;
        }
        else  {
            nxtL = *Lpp = L2;
            L2 = L2->sopc_next;
        }
        Lpp = &(nxtL->sopc_next);
        *Lpp = NULL;
    }
    PUSH = newL;
    return(TRUE);
} 
  

/*
 *  Push a null entry onto the stack, to begin new list.
 */
boolean doNull()
    {
    PUSH = NULL;
    return(TRUE);
}


/*
 *  Initialize for new ATTACH
 */
boolean Init_attach()
    {
    inv_stkptr = 0;
    deinitdevice();  /* Turn off packet filter */
    return(TRUE);
}


/* 
 *  Top-level of ATTACH.  
 *     (Bottom of) STACK contains ptr to beginning of chain of SOPCs.
 *     Append each SOPC to end of appropriate field chain.
 */
boolean Install_it()
    {
    register SOPC *sopcp, *tsopcp, **tpp;
    
    if (inv_stkptr != 1) {
        fprintf(stderr, "Stack Synch Error!\n");
        return(FALSE);
    }
    sopcp = POP;
    while (tsopcp = sopcp) {
        sopcp = sopcp->sopc_next ;
        tpp = &Fields[tsopcp->sopc_AtFIno].field_opcs ;
        while (*tpp)  tpp = &(*tpp)->sopc_next;
        *tpp = tsopcp ;
        tsopcp->sopc_next = NULL;
    }
    ComputeCuminv();  /* Compute cumulative invocation count */
    SetEtherType();   /* Try to narrow NIT3.x interface to specific
                       *   Ethernet type.
                       */
    return(TRUE);
}

/*
 *  Compare two fields (ie two nodes in parse tree) for ancestor
 *  relationship.
 *
 *    LEQtree(SOPC1, SOPC2) =>
 *         +1 if SOPC[f1] is ancestor of SOPC[f2] or is the same.
 *          0 Not (either f2 is ancestor of f1, or they are cousins)
 *         -1 One or both fields is psuedo, cannot be compared.
 *
 */
LEQtree(sp1, sp2)
    SOPC *sp1, *sp2;
    {
    register int FIno1 = sp1->sopc_AtFIno,
                 FIno2 = sp2->sopc_AtFIno, f;
    
    if (Fields[FIno1].field_parent == -1 || 
        Fields[FIno1].field_parent == -1 )
            return(-1);  /* Funny pseudo-field ... cannot be compared */
    if (FIno1 > FIno2) return(0);
        /* Starting at FIno2, search up parse tree for FIno1. */
    f =  FIno2;
    while (f != FIno1 && f != 0) f = Fields[f].field_parent;
    return((f)? 1 : 0);
}   
    
/*
 *  Routine to cleanup after Attach error.
 *
 *  Detach all invocations in each of
 *     n lists popped from top of stack.  If n is zero, clear entire stack.
 */
CleanUp(n)
    int n;
    {   
    SOPC *sopcp;
    
    while (inv_stkptr) {
        sopcp = POP;
        tree_prune((Xdebug)?0:1, NULL, 0, &sopcp, "*");
        if (n == 0) continue;
        else if (n == 1) return;
        else n--;
    }
}
    
    
    /***************************************************************
     * 
     *     ROUTINES TO CREATE, ATTACH, AND DELETE OBJECTS
     *
     *         bool Attach_SOBJ();  SOPC *Get_SOPC()
     *
     ***************************************************************/
/*
 *
 *    boolean Attach_SOBJ( code, &(struct invokes), parmlistp)
 *
 *  Find/create object and set up invocation of it from specified field.
 *  
 *  Return FALSE if an error occurs, else return TRUE.
 */
boolean Attach_SOBJ( code, invokesp, parmlistp)
    int code;
    register struct invokes *invokesp;
    char  *parmlistp;
    {
    char  *Objname = invokesp->inv_objname;
    register FIELD *FIptr1, *FIptr2 = NULL ;
    register SOBJ *sobjp ;
    SOPC  *sopcp;
    FIELD *find_field();
    SOBJ  *Get_SOBJ();
    int    ncases;
    extern int is_filter[];
    
    if ((FIptr1 = find_field(invokesp->inv_field1)) == NULL) { 
        attach_error("Bad field name", invokesp->inv_field1);
        return(FALSE);
    }       
    if (*invokesp->inv_field2 && 
            (FIptr2 = find_field(invokesp->inv_field2)) == NULL) {
        attach_error("Bad field name", invokesp->inv_field2);
        return(FALSE);
    }   
    if (!strcmp(Objname, NONAME)) {
        attach_error( "Illegal name", Objname);
        return(FALSE);
    }
        /* Create new object */
    if ((sobjp = Get_SOBJ(invokesp->inv_classno, Objname, 
                   parmlistp, invokesp->inv_nparms,
                   FIptr1->field_maxlen, 
                   (FIptr2)? FIptr2->field_maxlen: 0)) == NULL) {
        return(FALSE);  
    }
        /* Construct Invocation of object */
    ncases = (is_filter[sobjp->sob_class])? 2:0;
    if (sopcp = Get_SOPC( FIptr1, FIptr2, sobjp, Objname, ncases))  {   
        if (Xdebug)
            printf(
        "Attached: field(s) %s %s Obj %s Class %s: SOPC@%8.8x SOBJ@%8.8x\n", 
               invokesp->inv_field1,
               invokesp->inv_field2,
               invokesp->inv_objname,
               class_name[invokesp->inv_classno],
               sopcp, sobjp);
               
        PUSH = sopcp;  /* Push onto top of stack */
        return(TRUE);
    }
    Unget_SOBJ(sobjp);
    free(sopcp);
    return(FALSE);  
} /* Attach_SOBJ() */


/* 
 *   SOBJ  *Get_SOBJ(classname, object-name, parmlist, nparms,
 *                                                datalen1, datalen2 )
 *
 *     Find/Create specified object. 
 *
 *
 */
SOBJ *Get_SOBJ(Classno, objname, parmlistp, nparms, datalen, datalen2)
    char *objname;
    u_long *parmlistp ;
    int Classno, datalen, datalen2, nparms;
    {
    register SOBJ *sobjp ;
    GENERICOP *genopp = GenOpPtr[Classno-1] ;
    int n  = (nparms < MAXSVNPARM)? nparms : MAXSVNPARM;
    char *savevec();
    
    if (objname[0] != '\0') {
           /* Look for existing object */
        for (sobjp = SOBJ_list; sobjp; sobjp = sobjp->sob_next ) {
            if (!strcmp(objname, sobjp->sob_name)) {
                    /* FOUND IT */
                if ((Classno)&& Classno != sobjp->sob_class) {
                    attach_error("Class Conflict for", objname);
                    return(NULL) ;
                } 
                if ((nparms) &&
                    /* New invocation has parm list. It must match parm
                     *   list of existing object!
                     */
                        ((n != sobjp->sob_nparms) || 0 !=
                    	    bcmp((char *) parmlistp, sobjp->sob_parmp, 
                    	                                n*sizeof(u_long)))) {
                    attach_error("Parm list conflict for", objname);
                    return(NULL);
                }
                return(sobjp);  /* Use existing object */
            } 
        }
    }
    else objname = NONAME;
    
        /* New object. Must know its class... */
    if (Classno == 0) {
        attach_error("Unknown class for new object", objname) ;
        return(FALSE);
    } 
    if ((parmlistp = (u_long *)savevec(parmlistp,n*sizeof(u_long))) == NULL) {
    	attach_error("Storage shortage for", objname);
    	return(NULL);
    }  
        /* Call make entry point for class */
    sobjp = (*(genopp->make_xx))( objname, parmlistp, nparms,
                                           datalen, datalen2 ) ;
    if (sobjp == NULL) return( NULL ) ;
    sobjp->sob_name = savestring(objname) ;
    sobjp->sob_class = Classno ;
    sobjp->sob_nparms = nparms;
    sobjp->sob_parmp = (char *) parmlistp;
    sobjp->sob_next = SOBJ_list ;
    SOBJ_list = sobjp ;
    sobjp->sob_maketime = sobjp->sob_cleartime = CurrTime ;
    bcopy(genopp, &sobjp->sob_genop, sizeof(GENERICOP)) ;
    sobjp->sob_dleng = datalen;
    sobjp->sob_dleng2 = datalen2;
    sobjp->sob_totalno = sobjp->sob_orphans = 0;
    sobjp->sob_opccnt = sobjp->sob_lock = sobjp->sob_flags = 0;
    return(sobjp) ;
}

/*
 *      Unget_SOBJ( (SOBJ *) sobjp)
 */
Unget_SOBJ( sobjp)
    register SOBJ *sobjp;
    {
    register SOBJ **top = &SOBJ_list ;
    
            /* Must unchain from all-SOBJ list... */
    while (*top != sobjp) {
        if (*top == NULL) break;  /* SHOULD NEVER HAPPEN */
        top = &(*top)->sob_next ;
    }
    *top = (*top)->sob_next ;
    free(sobjp->sob_name) ; 
    (*(sobjp->sob_genop.delete_xx))(sobjp) ;
}

 /*  
  *   Get_SOPC(&field1, &field2, &SOBJ, objname, ncases)
  *
  *      Create invocation of object SOBJ on field(s) field1 [,field2].
  *      (objname included for error messages)
  */     
SOPC *Get_SOPC(FIptr1, FIptr2, sobjp, Objname, ncases)
    FIELD *FIptr1, *FIptr2;
    SOBJ *sobjp;
    char *Objname;
    int   ncases;
    {
    register SOPC *sopcp, **tp;
    int maxlen2 = (FIptr2)? FIptr2->field_maxlen: 0;
        
        /* Allocate space for an SOPC */
    if ((sopcp = (SOPC *) malloc(sizeof(SOPC)+(ncases-2)*sizeof(SOPC *))) ==
                                                                  NULL)  {
        SOBJ_error = "ATTACH: Severe memory shortage" ;
        return(NULL);
    }   
    sopcp->sopc_sobjp = sobjp ;
    sopcp->sopc_ncases =  ncases;
    tp = &sopcp->sopc_sublist[0];
    while (ncases-- > 0) *tp++ = NULL;
    sopcp->sopc_next = sopcp->sopc_indir = NULL;    

    if (sobjp->sob_opccnt > 0 && (
             sobjp->sob_dtype != FIptr1->field_dtype ||
             ((FIptr2) && sobjp->sob_dtype2 != FIptr2->field_dtype)))
        {
        attach_error( "Conflicting data type",  Objname) ;
    }
    else if (sobjp->sob_opccnt > 0 && (
             sobjp->sob_dleng != FIptr1->field_maxlen ||
             sobjp->sob_dleng2 != maxlen2 )) 
        {
        attach_error( "Conflicting field size",  Objname) ;
    }
    else {
        /* Everything is OK... */
        sopcp->sopc_FIptr1 =  FIptr1;  
        sopcp->sopc_FIptr2 =  FIptr2;  
        sopcp->sopc_AtFIno =
             (FIptr2 == NULL|| FIptr1->field_fino > FIptr2->field_fino) ? 
                  FIptr1->field_fino : FIptr2->field_fino ;
        sobjp->sob_dtype = FIptr1->field_dtype; 
        if (FIptr2) {
            sobjp->sob_dtype2 = FIptr2->field_dtype ;
        }
        sobjp->sob_opccnt++ ;  /* Count GETs on this object */
        return(sopcp);  /* OK */
    }
    free(sopcp);
    return(NULL); /* Failed */
} /* SOPC *Get_SOPC() */


         
/*
 *  Delete SOPC.
 *      If this was last call, delete SOBJ; 
 *      in any case, delete SOPC
 */     
DeleteSOPC(sopcp)
    register SOPC *sopcp;
    {
    SOBJ *sobjp = sopcp->sopc_sobjp;
    
    if (--(sobjp->sob_opccnt) == 0) 
        Unget_SOBJ(sobjp); 
    free(sopcp);
}   
    

char *savevec(cp, n)
    char *cp;
    int n;
    {
    char *sp;
    
    if (sp = (char *) malloc(n))
        bcopy(cp, sp, n);
    return(sp);
}

 
 /*  ComputeCuminv
  *     Compute for each node of the Field Parse Tree (FPT) the 
  *     cumulative invocation count, i.e., the number of nodes
  *     at or below the present node to which invocations attached.
  */
ComputeCuminv()
    {
    register int i;
    register FIELD *Fp;
    
    for (Fp = &Fields[1]; Fp->field_fino; Fp++) Fp->field_cuminv = 0;
    
    for (Fp = &Fields[1]; Fp->field_fino; Fp++)
        if (Fp->field_opcs) {
            Fp->field_cuminv++;
            i = Fp->field_fino;
            while ((i = Fields[i].field_parent) > 0)
	    	Fields[i].field_cuminv++;
        }
}
            
