


/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/ 
            
/*                      cmds.c
 *
 *  This file contains the routines to parse user commands.  Used by
 *  statspy to parse local commands, and by rspy to parse commands to be
 *  sent to remote statspy. 
 *
 */
 
/* CHANGES:
 *   29Nov88 ISI: Truncate excessively long names
 *   13Oct89 Merit: Add access control command: 'restrict'
 *   17Oct89 AKS/RTB: Add 'subnet' command
 *
 */  
  
#include <stdio.h>
#include <netdb.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <rpc/rpc.h>

#include "stat.h"
#include "scan.h"
    
extern char *field_name[];
extern Read_SOBJ(), Clear_SOBJ(), Attach_SOBJ(), Detach_SOBJ();
extern Show_Fields(), Show_restrict(), Do_Subnet();
extern int  Xdebug;
extern long CurrTime;
extern char *SOBJ_error;

#define ERR  -1 
#define ISTTY(x) ((x) == NULL)  
#define PARMALIGN(n)  ((n) > sizeof(long))? 0: sizeof(long)-(n) 
    
u_long Vals[MAXPARMS*LONGPERVAL];
char   Field1[MAX_FLDNAME], Field2[MAX_FLDNAME];
char   ObjName[MAX_OBJNAME];
extern char *strtcpy();


#define MAX_ENUMS 1024  /* Max size list of enum object specs */

    /* 
     *    Command scanner -- scan a single command read from stdin,
     *             and then return.
     * 
     *    fpin = file pointer for command input
     *    xpout = file handle for output
     *      In rspy:    XDR write handle.
     *      In statspy: NULL (output to local console).
     *
     */
Scan_Cmd(fpin, xpout)
    FILE *fpin, *xpout;
    {
    char *rp;
    char Ebuf[MAX_ENUMS];
    int interval, i, LastToken;
    u_long addr, mask, token;
    u_char addrmask[MAX_SUBNET_PARM];
    
    yyin = fpin;   /* Set LEX file pointer */
    LastToken = LEXCALL;
    while (1) {
        CurrTime = time(0);
                        
        switch (LastToken) {
        
        case EOFTOKEN:
            return(CMD_EOF);
                    
        case ';':
            return(CMD_OK);
            
        case READ:
            /* read [ <title-spec>  | ? ]   
             */
            if ((LastToken = LEXCALL) == '?') 
                ObjName[0] = '\0';  /* read object names (&attributes) */
            else if (LastToken == IDTOKEN) 
                strtcpy(ObjName, yytext, MAX_OBJNAME);
            else break;
            if (!Read_SOBJ(xpout, ObjName, 0))  XDRerr();
            return(CMD_OK);                     
                        
        case CLEAR:
            /* clear [ <title-spec> ] 
             */
            if ((LastToken = LEXCALL) == IDTOKEN) 
                strtcpy(ObjName, yytext, MAX_OBJNAME);
            else break;
            if (!Clear_SOBJ(xpout, ObjName)) XDRerr(); 
            return(CMD_OK);
                        
        case READCL:
            /* readclear [ <title-spec>  ]   
             */
            if ((LastToken = LEXCALL) == IDTOKEN) {
                strtcpy(ObjName, yytext, MAX_OBJNAME);
            }
            else break;
            if (!Read_SOBJ(xpout, ObjName, 1)) XDRerr();
            return(CMD_OK);             
            
        case SHOW:
            /*  show [ <field-spec>  | ? ]
             */
            if ((LastToken = LEXCALL) == '?')  
                ObjName[0] = '\0';  /* show field names, only */
            else if (LastToken == IDTOKEN) 
                strtcpy(ObjName, yytext, MAX_OBJNAME);
            else break;
            if (!Show_Fields(xpout, ObjName)) XDRerr();
            return(CMD_OK);
                        
        case DETACH:
            /* detach [ <title-spec> ]  
             */
            if ((LastToken = LEXCALL) == IDTOKEN) {
                strtcpy(ObjName, yytext, MAX_OBJNAME);
                if (!Detach_SOBJ(xpout, ObjName)) XDRerr();
                return(CMD_OK);
            }
            break;
            
        case ATTACH:
            /* 
             *   ATTACH { <attach list> }
             *
             *   This operation is atomic... as we parse the <attach list>,
             *   we build list of invocations and create objects as needed,
             *   but we don't install these invocations until the whole
             *   thing completes successfully.
             *
             */
            if (LEXCALL != '{') {
                printf("Bad Attach parm: no '{'\n");
                return(CMD_ERR);
            }   
            if (!Init_attach()) {
                printf("%s\n", SOBJ_error);         
                return(CMD_ERR);
            }
            LastToken = scan_attlist();
            
            if (LastToken != '}') {
                if (LastToken != ERR)
                    SOBJ_error =  "Missing '}'";
                CleanUp(0);
                printf("%s\n", SOBJ_error);         
                return(CMD_ERR);
            }           
                /* Install the invocations now... */
            if (!Install_it()) {;           
                printf("%s\n", SOBJ_error);         
                return(CMD_ERR);
            }
            return(CMD_OK);
                
        case ENUM:
            /* 
             *   ENUM  { <enum list> } | ENUM ?
             */             
            if ((LastToken = LEXCALL) == '?') {
                Ebuf[0] = '\0';
                show_enums(Ebuf, sizeof(Ebuf));
                printf("%s\n", Ebuf);
                return(CMD_OK);
            }           
            if (LastToken != '{') {
                printf( "Bad Enum parm: no {\n");
                return(CMD_ERR);
            }
            if ((i = scan_enum(fpin)) == 0)  {
                printf("Unknown name: %s\n", yytext);
                return(CMD_ERR);
            }           
            else if (i < 0)  {
                printf("Syntax error at: %s\n", yytext);
                return(CMD_ERR);
            }   
            return(CMD_OK);
                    
        case RESTRICT:
            /*
             *  RESTRICT {READONLY | READWRITE} ipaddr ipmask
             *    or
             *  RESTRICT ?
             */
            token = LEXCALL;
            if (token == '?') {
                    /* Print restrict table */
                Show_restrict();
                return(CMD_OK);
            }
            if (token != READONLY && token != READWRITE) {
                printf("Improper Restrict mode: %s\n", yytext);
                return(CMD_ERR);
            }
            if (LEXCALL != IPADDR) {
                printf("Improper Restrict address: %s\n", yytext);
                return(CMD_ERR);
            }           
            IPcnv(yytext, &addr);
            LastToken = LEXCALL;
            if (LastToken != IPADDR && LastToken != NUMBER) {
                printf("Improper Restrict mask: %s\n", yytext);
                return(CMD_ERR);
            }
            VALcnv(LastToken, &mask);
            return(Do_restrict((token == READWRITE), addr, mask));
            
        case SUBNET:
            /*
             *  SUBNET <net addr>, <mask>
             *
             *  Add specified (<net addr>, <mask>) pair to subnet list.
             *  If an IP address matches <net addr>, the specified <mask>
             *  will be used to define the corresponding subnet pseudo-field,
             *  instead of the default network class.
             *
             *  SUBNET ?
             *
             */
            if ((LastToken = LEXCALL) == '?')
                    /* Show subnets currently defined */
                addrmask[0] = '\0';
            else {
                if (LastToken != IPADDR) {
                    printf("Improper net address: %s\n", yytext);
                    return(CMD_ERR);
                }           
                IPcnv(yytext, &addrmask[0]);
                LastToken = LEXCALL;
                if (LastToken != IPADDR && LastToken != NUMBER) {
                    printf("Improper subnet mask: %s\n", yytext);
                    return(CMD_ERR);
                }
                VALcnv(LastToken, &addrmask[4]);
            }
            if (!Do_Subnet(xpout, addrmask)) XDRerr();
            return(CMD_OK);
             
        case '?':   
        case HELP:
            Do_help();
            return(CMD_OK);
                    
        default:
            switch (*yytext) {
            
            case 'p':
                /* poll <time> <title-spec> 
                 *         -- issues 'read <title>' every <time> secs.
                 *
                 *         -- poll 0  turns off polling.
                 */
                
                if (LEXCALL == NUMBER) {
                    interval = strtol(yytext, &rp, 0);
                    if ((LastToken = LEXCALL) != IDTOKEN)
                        rp = NULL;
                    else
                        rp = yytext;
                    SetPoll(interval, rp);
                }
                else 
                    SetPoll(-1, NULL);
                return(CMD_OK);
        
            case 'q':
                /* quit
                */
                exit(0);
                
            case 'h':
                /* host (rspy only)
                 */
                LastToken = LEXCALL;
                if (LastToken != IDTOKEN && LastToken != IPADDR) break;
                Do_host(yytext);
                return(CMD_OK);
                            
            default:
                break;
                
            }  /* end of little switch */   
        };  /* end of big switch */
        
        printf("Syntax error at: %s\n", yytext);
        return(CMD_ERR);
    }
}

XDRerr()
    {
    printf("ERROR: TCP conn failed\n");
}

    /* 
     *  Scan <attach list> ::=
     *      <attach spec> | <attach list> <attach spec>
     *
     *  Returns last token scanned, or error token.
     *
     */
int scan_attlist()
    {
    int  Begin = 1;     
    int  LastToken = LEXCALL;
    
    if (!doNull()) return(ERR);  /* Start of list */
    while (1) { 
        switch(LastToken) {
                    
        case '{':
        case IF:
        case RECORD:
        case ';':
            Begin = 0;          
            if ((LastToken = scan_attspec(LastToken)) == ERR)  {
                return(ERR);
            }                           
                /* Append (list of) SOPC's to previous list... */
            if (!doAppend()) return(ERR);
            break;
            
        default:
            if (Begin) {
                attach_error("Cannot start with", yytext);
                return(ERR);
            }
            return(LastToken);
        }
    }
}
    
    /* 
     *  Scan <attach spec> ::= 
     *     record <invoke spec> ; |
     *     if <invoke spec> <attach list> |
     *     if <invoke spec> <attach list> else <attach list>  |
     *     { <attach list> } | ;
     *
     *    Last token is passed as parameter, and routine returns last
     *    token as result.  This provides one-token lookahead.
     */
int scan_attspec(Token)
    int Token;
    {
    int   ifnot;
    struct invokes Invokes;
    
    if (Xdebug)
        printf("ATTACH SPEC token = %d text = %s\n", Token, yytext);
            
    switch (Token ) {
    
    case  ';':
        if (!doNull()) return(ERR);
        return(LEXCALL);
                        
    case RECORD:
        if ((LastToken = scan_invspec(&ifnot, &Invokes)) == ERR) 
            return(ERR);
        if (!Attach_SOBJ(Record, &Invokes, Vals))  return(ERR);
        if (LastToken == ';') { 
            return(LEXCALL);
        } 
        break;
        
    case IF:            
        if ((LastToken = scan_invspec(&ifnot, &Invokes)) == ERR)   
            return(ERR);
        if (!Attach_SOBJ(If, &Invokes, Vals))  return(ERR);
        
        if ((LastToken = scan_attspec(LastToken)) == ERR)  
            return(ERR);   
                    
        if (LastToken == ELSE) {            
            if ((LastToken = scan_attspec(LEXCALL)) == ERR)  
                return(ERR);
        }   
        else  
            if (!doNull()) return(ERR);
        if (!doIf(ifnot)) return(ERR);
        return(LastToken);
            
    case '{':
        if ((LastToken = scan_attlist()) == ERR)  return(ERR);
        if (LastToken == '}')  
            return(LEXCALL);                
    }   
            
    attach_error("Syntax error at", yytext) ;       
    return(ERR);                
}

    /*
     *  Scan <invocation spec> ::=
     *      <fieldname>... {is|isnot} <objname> |
     *      <fieldname>... {is|isnot} [<objname>] <class> ( <parmlist>)
     *
     *     Parse invocation statement, and build specification structure
     *    *invokesp, as well as parameter array vals[] (external).
     *
     *     Returns last scanned token (or error token ERR).
     *     If ISNOT appears, set *parity to 1, else 0.
     */         
int scan_invspec(parity, invokesp) 
    int *parity;
    struct invokes *invokesp;
    {
    int i, j;
    int   enumh = 0;
    char  valp[MAX_DLENG];
    
    bzero(Vals, sizeof(Vals));
    ObjName[0] = '\0';  /* Assume no object name */ 
    *parity = 0;  /* Assume if (true)... */ 
    Field2[0] = '\0';  /* Assume no field 2 */
    invokesp->inv_classno = 0;  /* Assume no class */

    if ((LastToken = LEXCALL) != IDTOKEN) {
        SOBJ_error = "Missing Field name";
        return(ERR);
    } 
    strtcpy(Field1, yytext, MAX_FLDNAME);
    if ((LastToken = LEXCALL) == IDTOKEN) {
        strtcpy(Field2, yytext, MAX_FLDNAME);
        LastToken = LEXCALL;
    }
    if (LastToken == ISNOT) 
        *parity = 1;  /* if NOT (true)... */
    else if (LastToken != INIS) {
        attach_error("Syntax error at", yytext);
        return(ERR);
    }
    if ((LastToken  = LEXCALL) == IDTOKEN) {
        strtcpy(ObjName, yytext, MAX_OBJNAME);
        LastToken = LEXCALL;
    }
    else if (LastToken != CLASS) {
        SOBJ_error = "Undefined Object";
        return(ERR);
    }   
        
    i = 0;
    if (LastToken == CLASS) {
        invokesp->inv_classno = yyval.int_val;
        
        if ((LastToken = LEXCALL) == '(') {
                /* Parameter list in parens.  List may contain qouted 
                 *   strings, which refer to labels from some enumeration
                 *   belonging to this object.  First, look up enumeration.
                 */ 
            if (ObjName[0] != '\0')  enumh = ExistEnumObj(ObjName); 
                    
            while ((LastToken = LEXCALL) != ')') {
                if (LastToken == STRING) {
                    yytext[yyleng-1] = '\0'; /* strip off quotation */
                    if (enumh == NULL) {
                        attach_error("No matching enum for", yytext+1);
                        return(ERR);
                    }
                    j = GetEnumVal(enumh, yytext+1, valp);
                }
                else
                    j = VALcnv(LastToken, valp) ;
                if (j == 0) {
                    attach_error("Unknown name", yytext) ;
                    return(ERR);
                }
                else if (j < 0) {
                    attach_error("Syntax error at", yytext) ;
                    return(ERR);
                }
                Vals[i] = 0;                
                bcopy(valp, (char *) &Vals[i] + (PARMALIGN(j)), j);
                i += ((j+3)/sizeof(long)) ;
                if (i >= MAXPARMS) {
                    attach_error("Too many parm values for", ObjName);
                    return(ERR);
                }
            } ;
            LastToken = LEXCALL;                            
        }
    }
    strcpy(invokesp->inv_field1, Field1);
    strcpy(invokesp->inv_field2, Field2);
    strcpy(invokesp->inv_objname, ObjName);
    invokesp->inv_nparms = i;
    return(LastToken);
}    
    
            

        
