/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/
/*
	connect.c
*/

#include <stdio.h>
#include <rpc/rpc.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <netdb.h>

#include "collect.h"

int	sockno,	
      	port = 0;
char 	hoststring[20];
FILE 	*fpin, 
	*fpout;
extern FILE *tracefp;
extern XDR *readp, *writep;

void
appendTime(cp, when)
   char *cp;
   u_long when;
{
   struct tm *t;

   t = localtime(&when);
   sprintf(endof(cp),"%s at %2.2d%2.2d%2.2d.%2.2d%2.2d%2.2d",
   	      hoststring,
	      t->tm_hour, t->tm_min, t->tm_sec,
              t->tm_mon+1, t->tm_mday, t->tm_year);

} /* appendTime */							    

void
flush_xdr()
{ 
   fflush(fpout);
};

void
perrorOut(cp)
   char *cp;
{
   char temp[100];

   sprintf(temp, "collect -- %s ", cp);
   appendTime(temp, time(0));
   perror(temp);
   
} /* perrorOut */

boolean
openconn(inetaddr)
   struct in_addr *inetaddr;
{
   FILE *fdopen();
   char temp[100];
	
   struct sockaddr_in myctladdr, remote;	
   int len = sizeof (myctladdr);
   u_long hostnum;

   hostnum = ntohl(*(u_long *) inetaddr);
   sprintf(hoststring,"%d.%d.%d.%d",
   	   (hostnum>>24), ((hostnum>>16)&0xff), 
	   ((hostnum>>8)&0xff), (hostnum&0xff));

   if (tracefp) {
      strcpy(temp, "Connecting ");
      appendTime(temp, time(0));
      strcat(temp, "\n");
      fprintf(tracefp, temp);
      fflush(tracefp);
      }

   /* Set up the socket address structure */
   bzero((char *)&remote, sizeof(struct sockaddr_in));
   remote.sin_family = AF_INET;
   bcopy((char *) inetaddr, (char *) &remote.sin_addr, 
   	 sizeof(struct in_addr)); 
   remote.sin_port = htons((u_short)(port==0)? STATD_PORT: port);
   	
   if ((sockno = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      perrorOut("socket");
      return(FALSE);
      }
	   
   /* Open TCP connection...  */
   if (connect(sockno, (struct sockaddr *) &remote, sizeof(remote)) < 0) {
      perrorOut("connect");
      close(sockno);
      sockno = -1;
      return(FALSE);
      };
	    
   if (getsockname(sockno, (char *)&myctladdr, &len) < 0) {
      perrorOut("getsockname");
      close(sockno);
      sockno = -1;
      return(FALSE);
      }
	    
   fpin =  NULL ;
   if ( ((fpin = fdopen(sockno, "r")) == NULL)  ||
        ((fpout = fdopen(sockno, "w")) == NULL)) {
      perrorOut("fdopen") ;
      if (fpin) 
	 fclose(fpin);
      close(sockno);
      sockno = -1;
      return(FALSE);
      };
      
   xdrstdio_create(readp, fpin, XDR_DECODE); 
   xdrstdio_create(writep, fpout, XDR_ENCODE); 
   
   return(TRUE);
} /* openconn */

void
closeconn()
{
   xdr_destroy(readp);	/* close streams */
   xdr_destroy(writep);

   if (tracefp) {
      fprintf(tracefp,"Closing...\n");
      fflush(tracefp);
      }
	
   if (sockno < 0)
      return;  

   fclose(fpin);
   fclose(fpout);
      
   close(sockno);
   sockno = -1;
} /* closeconn */

void
netclose()
{
   char temp[100];

   strcpy(temp, "collect -- timed out ");
   appendTime(temp, time(0));
   fprintf(stderr, "%s\n", temp);
			  
   closeconn();
} /* netclose */

