/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/

/*
 * 	logfile.c
 */
 
/*
 *  CHANGES:
 *    13Oct89 Merit: Set mode of new files
 */

#include <stdio.h>
#include <rpc/rpc.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/file.h>
#include <netdb.h>

#include "collect.h"

extern FILE *tracefp;
extern boolean needlog;
extern int chkpnt, interval, clear;
extern XDR read_s;
extern OBJ *rspP;
int openmode = -1;

struct obj_desc *o_ptr; /* ptr to current log info */
FILE *logfp;
u_long now;

void
create_log(host, objptr, start)
   char *host;
   struct obj_desc *objptr;
   u_long start;
{
   struct tm *t;
   char *cp, tempstr[200];

   t = localtime(&start);
   
   strncpy(tempstr,host,15);
   if (isalpha(host[0]) && (strlen(host) > 6)) {
      if ((cp = index(tempstr, '.')) != NULL)
         *cp = '\0';
      }
   strcat(tempstr,"-");
   strcat(tempstr,objptr->sname);
   
   sprintf(objptr->fname,"%s.%2.2d%2.2d.%2.2d%2.2d",
	   tempstr, t->tm_mon+1, t->tm_mday, t->tm_hour, t->tm_min);
   printf("\nCreating %s\n",objptr->fname);
   
   if ((objptr->logptr = fopen(objptr->fname,"w")) != NULL) {
      if (openmode != -1) fchmod(fileno(objptr->logptr), openmode);
      strcpy(tempstr, ctime(&now));
      tempstr[strlen(tempstr)-1] = 0;
      fprintf(objptr->logptr, "\nLog created on %s, for host %s.\n",
		tempstr, host);
      if (interval || chkpnt)
         fprintf(objptr->logptr, 
       "    Sample interval = %d minutes; checkpoint interval = %d minutes.\n",
	  	interval, chkpnt);
      if (clear)
         fprintf(objptr->logptr, "    Clear interval = %d minutes.\n", clear);
      fprintf(objptr->logptr, "    Object name = '%s'.\n\n", 
      	      objptr->sname);
      objptr->saventry = TRUE;
      objptr->logindex = ftell(objptr->logptr);
      }
} /* create_log */

struct obj_desc *
find_object(c_ptr, object)
   struct collector *c_ptr;
   char *object;
{
   struct obj_desc *objptr;
   int i;

   objptr = NULL;
   for (i=0; i < MAX_STATS; i++)
       if (!strlen(c_ptr->s[i].sname)) {
	  objptr = &c_ptr->s[i];
	  strcpy(objptr->sname, object);
	  create_log(c_ptr->hostname, objptr, c_ptr->creation);
	  break;
	  }
       else if (!strcmp(c_ptr->s[i].sname, object)) {
	  objptr = &c_ptr->s[i];
	  if (strlen(c_ptr->s[i].fname)) {
	     if (NULL == (objptr->logptr = fopen(objptr->fname, "r+")))
	        create_log(c_ptr->hostname, objptr, c_ptr->creation);
	     break;
	     }
	  else {
	     create_log(c_ptr->hostname, objptr, c_ptr->creation);
	     break;
             }
	  }
   return(objptr);
} /* find_object */


int
Start_Read(c_ptr, rspP)
   struct collector *c_ptr;
   OBJ *rspP;
{
   struct tm *tp;
        
   /* open log file */
   if (! needlog)
      logfp = stdout;
   else {
      if ((o_ptr = find_object(c_ptr, rspP->obj_name)) != NULL &&
          o_ptr->logptr != NULL)
         logfp = o_ptr->logptr;
      else
         return(FALSE);
            
   if (o_ptr->saventry) {
      /* ctime = creation - (creation mod 60 min) */
      tp = localtime(&c_ptr->creation);
      o_ptr->ctime = c_ptr->creation - ((tp->tm_min*60) + tp->tm_sec);
      if (chkpnt)
         while (rspP->obj_state.currtime >= (o_ptr->ctime + (60*chkpnt)))
            o_ptr->ctime += (60*chkpnt);
      o_ptr->rtime = rspP->obj_state.cleartime;
      }
   else {
      if (o_ptr->rtime != rspP->obj_state.cleartime) {
         fseek(logfp, 0, 2); 		/* start at end of log file */
         o_ptr->rtime = rspP->obj_state.cleartime;
	 fprintf(logfp, "\n");
         o_ptr->logindex = ftell(logfp);
         }
      else
         fseek(logfp,o_ptr->logindex,0); /* start at last start point */
      if (chkpnt && 
              (rspP->obj_state.currtime >= (o_ptr->ctime + (60*chkpnt)))) {
         o_ptr->saventry = TRUE;
         while (rspP->obj_state.currtime >= (o_ptr->ctime + (60*chkpnt)))
            o_ptr->ctime += (60*chkpnt);
         }
      }
   }
   return(TRUE);
} /* Read_logs */

void
log_reply(c_ptr)
   struct collector *c_ptr;
{  
   if (c_ptr) 
      now = c_ptr->readtime;
   logfp = NULL;
   Get_reply(&read_s, c_ptr);
}

void
End_Read()
{
    if (needlog) {
         if (o_ptr->saventry) {
            if (needlog) {
                fprintf(logfp, "\n");
                o_ptr->logindex = ftell(logfp);
                }
            o_ptr->saventry = FALSE;
            }
	 fprintf(logfp,"\n%c", EOF);
         fflush(logfp);
         fclose(logfp);
         logfp = NULL;
  	 }
} /* End_Read */
