 /************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/
 
/*     This file contains the semantic routines for collect to execute 
 *     remote attaches.
 *
 *            Init_attach()
 *            Install_it()
 *            doIf()
 *            doAppend()
 *            doNull()
 *            CleanUp()
 *            Attach_SOBJ()
 *            Detach_SOBJ()
 *            Show_Fields()
 *            Read_SOBJ()
 *            Clear_SOBJ()
 *            Do_Subnet()
 */

/*
 * CHANGES:
 *    17Oct89 AKS/RTB: Add Do_Subnet()
 *    25Oct89 ISI: Change some names
 *
 */ 
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <rpc/rpc.h>
#include "stat.h" 

extern XDR W_XDR, R_XDR;
char *SOBJ_error;
char  Error_msg[256];   
int   Flush();
char  out[256];
    

/*              int doIf()
 *
 *   Put a postfix IF/IFNOT operator into the XDR stream.
 *
 *   There are three implicit operands, all previously sent: the IF
 *   invocation, the TRUE alternative list, and the FALSE alternative
 *   list.  Returns TRUE if OK, else FALSE.
 *
 */
int doIf(isnot) 
    int  isnot;
    {
    int op = (isnot)? Ifnot:If;
    return(xdr_int(&W_XDR, &op));
}
    

/*         doAppend()
*
 *  Put an AppendList operator into the stream.
 */
boolean doAppend()
    {
    int op = AppendList;    
    return(xdr_int(&W_XDR, &op));
} 
  

/*
 *  Start new list.
 */
boolean doNull()
    {
    int op = StartList;
    return(xdr_int(&W_XDR, &op));
}


/*
 *  Initialize for new ATTACH
 */
boolean Init_attach()
    {
    int op = ATTACH_op;
    if (!Rconnect()) return(FALSE);
    return(xdr_int(&W_XDR, &op));   
}


/*
 *  Completion of ATTACH
 */
boolean Install_it()
    {
    int op = EndofReq;
    if (!xdr_int(&W_XDR, &op)) return(FALSE);
    Flush(&W_XDR);
    return(Get_reply(&R_XDR, NULL));    
    }

/*
 *  Routine to cleanup after Attach error.
 *
 *  Detach all invocations in each of
 *     n lists popped from top of stack.  If n is zero, clear entire stack.
 */
CleanUp(n)
    int n;
    {
    int op = Error ;    
    xdr_int(&W_XDR, &op);
    Flush(&W_XDR);
}


/*         int Attach_SOBJ(If/Record, struct invokes *,  char *)
 *
 */
boolean Attach_SOBJ(code, invokesp, parmlistp)
    int code;
    struct invokes *invokesp;
    char *parmlistp;
    {
    int invcode = (code==Record)? Record: IfInvoke;
    
    if (!xdr_int(&W_XDR,  &invcode)||
        !xdr_invoke(&W_XDR, invokesp, parmlistp)) {
        printf("Conn failed...\n");
        return(FALSE);
    };
    return(TRUE);       
}
    
        
/*        Detach_SOBJ(fh, objspec)
 */
boolean Detach_SOBJ(xdrs, objspec)
    XDR  *xdrs;
    char *objspec;
    {
    int code = DETACH_op;
    
    if (!Rconnect())  {
        printf("%s\n", SOBJ_error);
        return(FALSE);
    }
    if (!xdr_int(xdrs, &code) ||
        !xdr_string(xdrs, &objspec, MAX_OBJNAME)) {
        printf("Conn failed...\n");
        return(FALSE);
    }
    Flush(xdrs);
    return(Get_reply(&R_XDR, NULL));
}


/*        Read_SOBJ(fh, objspec, isclear)
 */
boolean Read_SOBJ(xdrs, objspec, isclear)
    XDR  *xdrs;
    char *objspec;
    int   isclear;
    {
    int code = (isclear)? READCL_op:READ_op;

    if (!Rconnect())  {
        printf("%s\n", SOBJ_error);
        return(FALSE);
    }
    if (!xdr_int(xdrs, &code) ||
        !xdr_string(xdrs, &objspec, MAX_OBJNAME)) {
        printf("Conn failed...\n");
        return(FALSE);
    }
    Flush(xdrs);
    return(Get_reply(&R_XDR, NULL));
}


/*        Clear_SOBJ(fh, objspec)
 */
boolean Clear_SOBJ(xdrs, objspec)
    XDR *xdrs;
    char *objspec;
    {
    int code = CLEAR_op;
    
    if (!Rconnect())  {
        printf("%s\n", SOBJ_error);
        return(FALSE);
    }
    if (!xdr_int(xdrs, &code) ||
        !xdr_string(xdrs, &objspec, MAX_OBJNAME)) {
        printf("Conn failed...\n");
        return(FALSE);
    };
    Flush(xdrs);
    return(Get_reply(&R_XDR, NULL));
}


/*        Show_Fields(fh, fldspec)
 */
boolean Show_Fields(xdrs, fldspec)
    XDR *xdrs;
    char *fldspec;
    {
    int code = SHOW_op;
    
    if (!Rconnect())  {
        printf("%s\n", SOBJ_error);
        return(FALSE);
    }
    if (!xdr_int(xdrs, &code) ||
        !xdr_string(xdrs, &fldspec, MAX_OBJNAME)) {
        printf("Conn failed...\n");
        return(FALSE);
    };
    Flush(xdrs);
    return(Get_reply(&R_XDR, NULL));
}


/*        Do_Subnet(fh, parm)
 */
boolean Do_Subnet(xdrs, addrmask)
    XDR *xdrs;
    u_char *addrmask;
    {
    int code = SUBNET_op;
    int len = (addrmask[0])? MAX_SUBNET_PARM:0;
    
    if (!Rconnect())  {
        printf("%s\n", SOBJ_error);
        return(FALSE);
    }
    if (!xdr_int(xdrs, &code) ||
        !xdr_bytes(xdrs, &addrmask, &len, MAX_SUBNET_PARM)) {
        printf("Conn failed...\n");
        return(FALSE);
    };
    Flush(xdrs);
    return(Get_reply(&R_XDR, NULL));
} /* Do_Subnet */
    
    /***********************************************************
     *
     *   UTILITY ROUTINES
     *
     ***********************************************************/   

Flush(fh)
int fh;
    {
    extern FILE *noutfp;
    
    if (fh == 0) {
        printf("\n");
        fflush(stdout);
    }
    else    
        fflush(noutfp);
}


attach_error(msgp, namep)
    char *msgp, *namep;
    {
    sprintf(SOBJ_error = Error_msg, "ATTACH error -- %s: %s", 
           msgp, namep);
}
    
