/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/

/*
 *	                  reply.c
 *
 *   Routines to format/log XDR replies to commands to remote statspy.
 *   Used by collect and by rspy.
 */

/*
 *  CHANGES:
 *    15Oct89 ISI: Return Boolean: Get_reply()
 *    17Oct89 ISI: Support remote "subnet ?" commmand
 *      Rel 2.4:
 *    01Nov89 ISI: Support MAB, MSB, FAB, FOB object classes; change calling
 *                        sequences for ReadFO(), MA_list(), and FAO_list().
 */
 
#include <stdio.h>
#include <rpc/rpc.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>

#include "collect.h"
#include "sobj.h"

#define endof(x) (x + strlen(x))
#define pad(x,col) while (strlen(x)<col) strcat(x, " ")

extern FILE *tracefp;
extern FILE *logfp;
u_long get_long();
OBJ  rsp_obj, *rspP = &rsp_obj;
char out[256];



u_long
get_long(xdrs, out_str, label)
   XDR *xdrs;
   char *out_str, *label;
{
   u_long i;

   if (!xdr_int(xdrs, &i))  {
      sprintf(out_str,"Bad xdr_int call in get_long");
      i = 0;
      }
   else {
      sprintf(out_str,"%s = %u\n", label, i);
      }
   return(i);
} /* get_long */

void
ma_list(xdrs)
XDR *xdrs;
{
   MADE  bin;
   int   j, count;
   boolean isEnum;

   isEnum = ExistEnumObj(rspP->obj_name);
   
   count = get_long(xdrs, out, "  #bins");
   log(out, logfp);
   
   MakeCols(NULL);
   for (j=0;j<count;j++) {
      if (!MA_list(xdrs, out, sizeof(out), &bin, &rspP->obj_state, isEnum))
         {
         printf(" BAD MA_list! ");
         break;
         }
      else
         log(out, logfp);
    }
} /* ma_list */
    
void
mfb_list(xdrs)
XDR *xdrs;
{
   MFBDE  bin;
   int   j, count;
   boolean isEnum;
   struct dblword  bytdbl;
   char bytbuf[18];

   isEnum = ExistEnumObj(rspP->obj_name);
   if (!xdr_opaque(xdrs, &bytdbl, sizeof(bytdbl))) {
       printf(" Bad MFB read");
       return;
   }
   print_dbl(bytbuf, bytdbl.H, bytdbl.L);
   sprintf(out, "  Total Bytes= %sB  ", bytbuf);  
   count = get_long(xdrs, endof(out), "  #bins");  
   log(out, logfp);
   
   MakeCols(NULL);
   for (j=0;j<count;j++) {
      if (!MFB_list(xdrs, out, sizeof(out), &bin, &rspP->obj_state, isEnum))
         {
         printf(" BAD MFB_list! ");
         break;
         }
      else
         log(out, logfp);
    }
} /* mfb_list */
    
void
fa_list(xdrs)
XDR *xdrs;
{
   FADE  bin;
   int   j, count;
   boolean isEnum;

   isEnum = ExistEnumObj(rspP->obj_name);
   
   count = get_long(xdrs, out, "  #bins");
   log(out, logfp);
   
   MakeCols(NULL);
   for (j=0;j<count;j++) {
      if (!FAO_list(xdrs, out, sizeof(out), &bin, &rspP->obj_state, 
                        isEnum, FAOBJ))
         {
         printf(" BAD FAO_list! ");
         break;
         }
      else
         log(out, logfp);
    }
} /* fa_list */
    
void
ws_list(xdrs)
XDR *xdrs;
{
   long   *p2bp = NULL;
   int    count = MAX_P2BINS; 
   u_long distinct;
   
   if (!ReadWS(xdrs, &p2bp, &count, &distinct, rspP, 1))
      printf(" BAD ReadWS! ");

} /* ws_list */

void
fo_list(xdrs)
XDR *xdrs;
{
   FODE  bins[MAX_FONO];  /* Place to decode data from bins */
   FODE *binp = bins;
   int   count = MAX_FONO; 
   long  defaultno; 
   
   if (!ReadFO(xdrs, &binp, &count, &defaultno, rspP, sizeof(FODE), 1, NULL))
      printf(" BAD ReadFO! ");
} /* fo_list */

void
fob_list(xdrs)
XDR *xdrs;
{
   FOBDE  bins[MAX_FONO];  /* Place to decode data from bins */
   FOBDE *binp = bins;
   int    count = MAX_FONO; 
   long   defaultno;
   struct FOparmB totalbytes; 
   
   if (!ReadFO(xdrs, &binp, &count, &defaultno, rspP, sizeof(FOBDE), 1,  
                                                                &totalbytes))
      printf(" BAD ReadFOB! ");
} /* fo_list */
                     
void
p2_list(xdrs)
XDR *xdrs;
{
   P2DATA p2data;
   long   *p2bp = NULL;
   int    count = MAX_P2BINS;
   
   if (!ReadP2(xdrs, &p2bp, &count, &p2data, 1))
      printf(" BAD ReadP2! ");
      
} /* p2_list */

void
hi_list(xdrs)
XDR *xdrs;
{
   P2DATA p2data;
   HistE  *binpp = NULL;
   int    count = MAX_HIBINS;

   if (!ReadHI(xdrs, &binpp, &count, &p2data, 1))
        printf(" BAD ReadHI! "); 
        
} /* hi_list */

void
bp_list(xdrs)
XDR *xdrs;
{
   int count, j;
   BPE bpe;
   
   count = get_long(xdrs, out, "  #packets saved");
   log(out, logfp);
   for (j=0;j<count;j++) {
       if (!ReadBP(xdrs, &bpe, rspP, 1))
          {
          printf(" BAD ReadBP! ");
          break;
          }
    }
    
} /* bp_list */

void
show_list(xdrs)
    XDR *xdrs;
    {
    char temp[MAX_FLDNAME];
    int  oper;
    int  level = 0;
    struct invokes Invokes;
    
    if (!SHOWfield(xdrs, temp, 0, 1)) return;
    while (1) {
    
        if (!xdr_int(xdrs, &oper)) return;
        switch (oper) {
    
        case EndofResp:
            return;
            
        case SHOW_op:  /* Start new field */
            if (!SHOWfield(xdrs, temp, 0, 1)) return;
            level = 0;
            break;
        
        case If: 
        case Record:        
            if (!SHOWinv(xdrs, &Invokes, level, 1)) return;
            break;
        
        case StartTsubl:
            SHOWsublist(NULL, TRUE, ++level, 1);
            break;
        
        case StartFsubl:
            SHOWsublist(NULL, FALSE, ++level, 1);
            break;
        
        case EndofSubl:
            level--;
            break;
        
        default:
            printf("SHOW op?? %d\n", oper);
            return;
        }
    }
} /* show_list */
               
/* 
 *  Decode reply from XDR and log/print results
 *      xdrs -- XDR * to be decoded.
 *      cptr -- collector * to be passed to Start_Read(cptr, rspH) 
 *			when a Read is to be decoded.
 */   
boolean
Get_reply(xdrs, cptr) 
    XDR *xdrs;
    struct collector *cptr;
    {
    int Resp_code, Resp_op;
    char *temp = NULL; /* Let XDR malloc areas for strngs... */
    int Datalen;
   
    if (!xdr_int(xdrs, &Resp_code)) return(FALSE);
    switch (Resp_code) {
   
    case RESP_ERROR:
        log("Error --", tracefp);
        if (xdr_string(xdrs, &temp, 256)) {
            log(temp, tracefp);
            free(temp);
        }
        log("\n", tracefp);
        return(TRUE);
         
    case RESP_OK:
        log("OK.\n", tracefp);
        while (1) {  /* Allow for multiple Response ops (Read) */
            if (!xdr_int(xdrs, &Resp_op)) return(FALSE);
            switch (Resp_op) {
         
            case EndofResp: 
                /****log("\nEnd of Response\n", tracefp);***/
                log("\n", tracefp);
                return(TRUE);
             
            case CLEAR_op:
            case ATTACH_op:
            case DETACH_op:
            case SHOWnames_op:
            case SUBNET_op:
                if (xdr_string(xdrs, &temp, 8000)) {
                    log(temp, tracefp);
                    free(temp);
                }
                log("\n", tracefp);
                return(TRUE);
                                                      
            case READ_op:
            case READCL_op:
                out[0] = '\0';
                if (!ReadStd(xdrs, out, 256, rspP)) return(FALSE);             
                if (!Start_Read(cptr, rspP)) return(FALSE); 
                log(out, logfp);
                if ((Datalen = rspP->obj_totlength  - sizeof(struct Obj_state) 
                                        - 4 /* sizeof(class)*/) <= 0) 
                    break;  /* Read ? */

                switch (rspP->obj_class) {
            
                case FOOBJ: 
                    fo_list(xdrs);
                    break;
                case HIOBJ: 
                    hi_list(xdrs);
                    break;
                case P2OBJ: 
                    p2_list(xdrs);
                     break;
                case WSOBJ: 
                    ws_list(xdrs);
                    break;
                case FAOBJ: 
                case FA2OBJ: 
                    fa_list(xdrs);
                    break;
                case MSOBJ:
                case MAOBJ: 
                case MA2OBJ: 
                    ma_list(xdrs);
                    break;
                case MABOBJ:
                case MSBOBJ:
                case FABOBJ:
                   mfb_list(xdrs);
                    break;
                case FOBOBJ:
                    fob_list(xdrs);
                    break;
                case BPOBJ: 
                    bp_list(xdrs);
                    break;
                case SFOBJ:
                case RFOBJ:
                case EQOBJ:
                    get_long(xdrs, out, "  #TRUE");
                    log(out, logfp);
                    break;
                default: 
                    hex_dump(xdrs, logfp, 
                        out, sizeof(out), Datalen/4, DTYP_bits);
                } /* End switch on Read class */
                log("\n",tracefp);
                End_Read();
                break;  /* Back for another READ segment, or EndofResp */
            
            case SHOW_op:
                show_list(xdrs); 
                log("\n", logfp);
                return(TRUE);   

             default: 
                if (tracefp) fprintf(tracefp,"op = %xx??\n\n",Resp_op);
                return(TRUE);
            }  /* End of while(1) loop */
        }  /* End switch on Resp_op */
        break;
            
    default:
        if (tracefp) fprintf(tracefp,"code = %xx\??n",Resp_code);
    } /* End switch on Resp_code */
    return(FALSE);
} /* Get_reply */


