
/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   * 
 *     as this notice is not removed.                                     *
 **************************************************************************/ 

       /* 
        *               stat.h
        *  Define external interface of statspy
        *
        */
 
 /* CHANGES:
  *   17Nov88 ISI: Add eqf object class 
  *   29Nov88 ISI: Larger MAX_OBJNAME, MAX_LABELS
  *   17Oct89 AKS/ISI: add subnets, other ops.
  *      Rel 2.4:
  *   01Nov89 ISI: Add MAB, MSB, FAB, and FOB objects.
  *   30Mov89 ISI: Add VERSION_op.
  */
 
#define STATD_PORT  2222   /* Default Port */
#define WORDS(L)   ((L+3)>>2)
typedef int boolean;

    /* Request & Response Operation Codes
     */ 
#define EndofReq   0
#define EndofResp  0
#define ATTACH_op  1
#define DETACH_op  2
#define READ_op    3
#define CLEAR_op   4
#define READCL_op  5
#define SHOW_op    6
#define SHOWnames_op  7
#define SUBNET_op  8
#define Error      9
  /* Following are used for ATTACH parm strings */
#define Record     10
#define If         11
#define Ifnot      12
#define IfInvoke   13
#define StartList  14
#define AppendList 15
#define And        16
#define Or         17
#define SymIf      18
#define SymIfNot   19
  /* Following are used for SHOW replies */
#define StartTsubl 20
#define StartFsubl 21 
#define EndofSubl  22
#define VERSION_op 23   /* Remote inquiry of version */


    /* Encoding of object classes 
     *  (MUST AGREE with declarations:
     *     char * class_name[] in scan.c
     *     char * class_nick[] in scan.c
     *     GENERICOP *GenOpPtr[] in attach.c)
     */
#define FOOBJ  1   /* freq-only */
#define HIOBJ  2   /* hist */
#define P2OBJ  3   /* hist-pwr2 */
#define FAOBJ  4   /* freq-all */
#define BPOBJ  5   /* bin-pkt */
#define FA2OBJ 6   /* freq-all2 */
#define RFOBJ  7   /* rangef */
#define SFOBJ  8   /* setf */
#define MAOBJ  9   /* matrix-all */
#define WSOBJ  10  /* working-set */
#define MSOBJ  11  /* matrix-sym */
#define MA2OBJ 12  /* matrix-all */
#define EQOBJ  13  /* eqf */
#define FABOBJ 14  /* freq-all-bytes */ 
#define FOBOBJ 15  /* freq-only-bytes */ 
#define MABOBJ 16  /* matrix-all-bytes */
#define MSBOBJ 17  /* matrix-sym-bytes */

    /* Maximum  sizes... */
#define MAX_LABELS  128
#define MAX_OBJNAME  64
#define MAX_FLDNAME  16
#define MAX_CLSNAME  16
#define MAX_SUBNET_PARM 8
#define MAX_DLENG    8  /* Max length of a field value */
#define MAXPARMS   256  /* Maximum number of paramete vals to object */
#define LONGPERVAL  ((MAX_DLENG+3)/4)

   /* Used within STATSPY: */
#define CMD_OK    0
#define CMD_EOF  -1
#define CMD_ERR  -2
#define CMD_QUIT -3


     
struct Obj_state  {  /* Standard object state information */
    long  currtime,
            maketime,
            cleartime;
    
    u_long  totalno ;
    u_long  orphans ;
        
    int  datatype;
        /* 
         *   Data types -- for printing values nicely
         */
#define DTYP_unknown   0    /* ?? */
#define DTYP_int       1    /* Dec integer */
#define DTYP_IP        2    /* 32-bit IP address (host order) */
#define DTYP_bits      3    /* Hex */
#define DTYP_EtherA    4    /* Ethernet Address */
#define DTYP_packet    5    /* Packet   */
    
    int  dleng ;
    int  datatyp2;
    int  dleng2;
};   

    /*
     *  Define external data common to all object classes
     */
#define OBJ struct object_rep
OBJ {
    char   obj_name[MAX_OBJNAME];  /* Name of object */
    int    obj_totlength;  /* Total length of the 
                              rest of this response segment */
    struct Obj_state obj_state;
    int    obj_class;       /* Class code */
};

    /*****************************************************************
     *
     *    XDR Description of a Request to statd 
     *
     *****************************************************************  
    
    union switch(Req_code) {      -- Request Code --
     
    case READ_op:
    case REACL_op:
    case CLEAR_op:
    case SHOW_op:
    case DETACH_op:
        string Request<> ;   -- Parm String --
        
    case SUBNET_op:  /* Parm string is empty for show subnets, else
                      * it is 8 bytes of address and mask (u_char).

        string Request<> ;   -- Parm String --
    
    case ATTACH_op:
        --   Generally, this is an (operator, operand) expression in postfix
         *   order.  Interpretation is straight-forward using a stack
         *   machine.
         *
         *   The operands are invocation specifications encoded in XDR as:
         *
        string Att_Field_name_1<>;
        string Att_Field_name_2<>;  -- zero-length for unary op --
        string Att_ObjName<>;       -- zero-length for unnamed object --
        int    Att_Classno;
         *
         *   with the following external representation: */
         
struct invokes {
    char  inv_field1[MAX_FLDNAME];  /* Name of Field 1 */
    char  inv_field2[MAX_FLDNAME];  /* Name of Field 2 */
    char  inv_objname[MAX_OBJNAME]; /* Name of object */
    int   inv_classno; /* Class number */
    int   inv_nparms;  /* Number long words in parameter value array */
} ;
        /*  The postfix operators are StartList, AppendList, If, 
         *    Ifnot, Record, or Error.
         */
            

    /*****************************************************************
     *
     *    XDR Description of a Response from statd 
     *
     *****************************************************************
        
    union switch(int Resp_code) {
    
    case 1:   -- Error --
        string Resp_Error<127> ;  
        
    case 0:   -- OK --
    
        union switch (int Resp_op) {   --  May be multiple repetitions of this
                                       --  for READ SHOW responses; after last
                                       --  repetition there will be Resp_op=0.          
        case 0:                 
                --  In this case, we depart from XDR... we allow any number
                --  of responses to the same READ or SHOW request, each
                --  beginning with the same Resp_op descriminant value;
                --  following the last one, union descriminant will be 0.           
        case DETACH_op:
            string Resp_comment<>;
                            
        case ATTACH_op:
            string Resp_comment<>;
                
        case CLEAR_op:
            string Resp_comment<>;
                                
        case SHOWnames_op:
            string Resp_comment<>;
                
        case READ_op:
            string Resp_objname<MAX_OBJNAME> ;  -- Object name --
            
            int   totlength;  -- Total length of the rest of
                                 this response segment --
            
            struct Obj_state ; -- standard prefix --
                
            union switch (int Resp_objclass)  { 
                 -- Select by Class of object --                  */
        
/***********************************************************************        
 *          case FOOBJ:    -- Frequency: Only --
 */
#define FOr_SIZE(N,L) (2 + (N)*(1+WORDS(L)))*sizeof(u_long)                 
#define MAX_FONO  256 /* Max number predefined FREQ_ONLY bins */
#define FODE struct only_dataelement

/*              u_long   FOr_other ;      */

                    /* FREQ_ONLY element */ 
                FODE {
                    u_long   fod_count; 
                    u_char   fod_value[MAX_DLENG];  /* Value for this bin */
    /* Note: the real representation of value is XDR opaque, ie byte string padded
     * to nearest fullword.
     */
                }       /* FOr_list<>  */ ;

            
/***********************************************************************        
 *          case FAOBJ:   -- Frequency: All --
 */             
#define FADE struct all_dataelement 
#define FAr_SIZE(N, L) (1 + (N)*(2+WORDS(L)))*sizeof(u_long)

                     /* FREQ_ALL data element */    
                FADE {
                    FODE   fad_fode;
                    long   fad_time;   /* Last Incr: Secs since Jan 1, 1970 */
                }       /*  FAr_list<> */ ;                 
            
#define fad_count fad_fode.fod_count
#define fad_value fad_fode.fod_value

            
/***********************************************************************        
 *          case MAOBJ:   -- Matrix: All --
 */             
#define MAr_SIZE(N, L)  (1 + (N)*(2+WORDS(L)))*sizeof(u_long)
#define MADE struct ma_dataelement 

                     /* matrix_all data element */  
                MADE {
                    u_long mad_count;
                    long mad_time;   /* Last Incr: Secs since Jan 1, 1970 */
                    u_char mad_value[2*MAX_DLENG];  /* Concatenation of
                                                       pair of values */
    /* Note: the real representation of value is XDR opaque, ie byte string padded
     * to nearest fullword.
     */
                }       /*  MAr_list<> */ ;                 

         
/***********************************************************************        
 *          case FOBOBJ:    -- Frequency Only - Bytes --
 */
#define FOBr_SIZE(N,L) (6 + (N)*(3+WORDS(L)))*sizeof(u_long)                 
#define FOBDE struct fob_dataelement

/*              u_long   FOBr_other ;     # default packets (not in any bin)
                u_long   FOBr_othHbytes;  Total bytes for default packets
                u_long   FOBr_othLbytes;
                u_long   FOBr_totHbytes;  Total bytes including defaults packets
                u_long   FOBr_totLbytes; 
    */

                    /* freq-only-bytes element */ 
                FOBDE {
                    u_long   fob_count;
                    u_long   fob_Hbyte;   /* Higher-order byte count */ 
                    u_long   fob_Lbyte;   /* Low-order 3 bytes of count */
                    u_char   fob_value[MAX_DLENG];  /* Value for this bin */
    /* Note: the real representation of value is XDR opaque, ie byte string padded
     * to nearest fullword.
     */
                }       /* FOBr_list<>  */ ;

 
/***********************************************************************        
 *          case FABOBJ:   -- Frequency All - Bytes --
 */             
       /* Same network representation as MABOBJ, following: */
            
/***********************************************************************        
 *          case MABOBJ:   -- Matrix All/Sym - Bytes --
 */             
#define MFBr_SIZE(N, L) (3+ (N)*(4+WORDS(L)))*sizeof(u_long)
#define MFBDE struct mfb_dataelement
 
          /*    u_long   MFBr_Hbytes;
                u_long   MFBr_Lbytes;      */

                     /* matrix_all data element */  
                MFBDE {
                    u_long mfbd_count;  /* Freq count */
                    long   mfbd_time;   /* Last Incr: Secs since Jan 1, 1970 */
                    u_long mfbd_Hbyte;  /* High-order byte count           */
                    u_long mfbd_Lbyte;  /* Low order 3 bytes of byte count */
                    u_char mfbd_value[2*MAX_DLENG];  /* Value(s) */
    /* Note: the real representation of value is XDR opaque, ie byte string padded
     * to nearest fullword.
     */
                }       /*  MFBr_list<> */ ;                 
                        
    
/***********************************************************************        
 *          case HIOBJ: -- Histogram --
 */
#define HistDATA struct hist_data
#define HIr_SIZE(n)  (1 + n)*sizeof(u_long)+sizeof(HistDATA)
#define MAX_HIBINS 1024
#define HistE u_long    /* Element (bin) */

                HistDATA {
                    u_long Hist_scalef ,
                           Hist_avg ,
                           Hist_min ,
                           Hist_max ,
                           Hist_default ;
                    };
                    
/*                  HistE HIr_count<MAX_HIBINS>;  -- bins -- 
 */

/***********************************************************************        
 *          case P2OBJ:   -- Histogram: Powers of 2 -- 
 */

#define P2DATA HistDATA
#define P2r_SIZE(n)   (1 + n)*sizeof(u_long)+sizeof(P2DATA)
#define MAX_P2BINS 32

/*              P2DATA ;                    
                HistE   P2r_count<MAX_P2BINS> ;  -- bins --
*/                          


/***********************************************************************        
 *          case WSOBJ:   -- Working Set -- Cache Misses for cache sizes
 *                              increasing as Powers of 2 -- 
 */

#define MAX_LOG2N 12
#define WSr_SIZE(N) (2+N)*sizeof(u_long)

/*              u_long   WSr_distinct;
                u_long   WSr_count<MAX_P2BINS> ;  -- bins --
*/                          

/***********************************************************************    
 *          case BPOBJ:  -- Binary Packets --
 *
 *   NOTE:  The Resp_std.datatype field is meaningless for this class.
 *    The entire packet header is included, for display in simple hex.
 */
                             
#define BPE struct bpe_entry 
#define BPr_SIZE(n, l) (1+2*(n))*sizeof(u_long)+l
#define MAX_BPSIZE 63 /* Maximum bytes in BP packet */

                BPE {
                    long bpe_time ;   /* Timestamp when collected */
                    u_char bpe_pktlen ; /* Actual length saved in
                                                     following slot */
                    char   bpe_packet[MAX_BPSIZE] ; /* Packet saved */ 
                } ;
                

/***********************************************************************        
 *          case SFOBJ:  -- Set Filter --
 *
                u_long SFr_trueno;
 */
#define SFr_SIZE  sizeof(u_long)


/***********************************************************************    
 *          case RFOBJ:  -- Range Filter --
 *
                u_long RFr_trueno;
 */
#define RFr_SIZE  sizeof(u_long)


/***********************************************************************
             } -- End of switch on Resp_objclass and READ response  --
             
 ***********************************************************************
          
        case SHOW_op:
                
            string Resp_fieldname<>;  -- Field AT which evaluation happens --
            
                   --  Generally, there follows an operator-prefix 
                   *   description of the interpretive execution trees.
                   *   Operands are invocations as defined earlier as
                   *   struct invokes under ATTACH. Operators are If, Record.
                   *   Each sublist is delimited by (StartTlist, EndofResp)
                   *   pair or (StartFlist, EndofResp) pair.                   
                 
        }  -- End of switch on Resp_op --
            
    } -- End of OK response --
                 
****/               
    
#define RESP_OK 0
#define RESP_ERROR 1
     
#define endof(x) (x + strlen(x))
#define log(s,x) {if (x) fprintf(x,s);}
extern FILE *logfp, *tracefp;

#define NONAME  "(no name)"
#define INFINITY 0x7FFFFFFF
