

/************************************************************************** 
 *     NNstat -- Internet Statistics Collection Package                   *
 *                                                                        *
 *        Written by: Bob Braden & Annette DeSchon                        *
 *        USC Information Sciences Institute                              *
 *        Marina del Rey, California                                      *
 *        1987-1988                                                       *
 *                                                                        *
 *     NNStat is Public Domain, and may be used for any purpose as long   *
 *     as this notice is not removed.                                     *
 **************************************************************************/ 

/*                           statio.c
 * 
 *     This file contains the common network I/O and display routines
 *     used by statspy, rspy, and collect.
 */

/* CHANGES:
 *   19Nov88 ISI:  tweak SHOW format.
 *    5Sep89 LANL:  Fix SPARC problem in print_val.
 *   13Oct89 Merit: Add option to print time in GMT.
 *   14Dec89 ISI:   Fix bug in print_val.
 *      Rel 2.4:
 *    1Nov89 ISI: Add code for new objects FAB, FOB, MAB, MSB.
 *                Modify calling seqs of MA_list, FAO_list, MFB_list, ReadFO.
 *   30Nov89 Merit/ISI: Terse mode changes. 
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <rpc/rpc.h>
#include <netdb.h>
#include "stat.h"
#include "sobj.h"

extern int asock ; 
char *Get() ;
extern XDR W_XDR;

#define MAXCOLS 80
int   MaxItemSize = 0, LineUsed = 0, prevW = 0;
#define Align4(x)  (3 - ((strlen(x)+3)&3))
int  aligncnt;

boolean print_utime = 0;
boolean tersemode = 0;

extern char *class_name[];
extern char *ether_ntoa();

void MakeCols();
void print_delta();
void print_time();
void hex_dump();
void hex_print();
void print_val();
void print_percent();
void print_dbl();


/********************************************************************
 *
 *                      Read I/O ROUTINES
 *
 *   These routines are used by both statspy and collect to implement
 *   Read operations on particular object classes.  Each routine may be 
 *   called to perform any of the following functions:
 *
 *     In collect -- decode the data from XDR and format a display of it.
 *     In statspy (local read) -- format a display of the given data. 
 *     In statspy (remote read) -- encode the given data into XDR. 
 *
 *   Each routine returns FALSE if any XDR routine returns FALSE,
 *   otherwise it returns TRUE.
 *
 *
 ********************************************************************/
 
    /*
     *  ReadStd: Encode/decode and/or display the standard contents of
     *  a statistical object.
     *   */

boolean ReadStd(xdrs, outp, outl, objp)
    register XDR *xdrs;
    register char *outp;
    int outl;
    register OBJ *objp;
{
    int n;
    char *cp = objp->obj_name;
    
    if (xdrs) {
       if (!xdr_string(xdrs, &cp, MAX_OBJNAME) ||
           !xdr_int(xdrs, &objp->obj_totlength) ||
           !xdr_long(xdrs, &objp->obj_state.currtime) ||
           !xdr_long(xdrs, &objp->obj_state.maketime) ||
           !xdr_long(xdrs, &objp->obj_state.cleartime) ||
           !xdr_u_long(xdrs, &objp->obj_state.totalno) ||
           !xdr_u_long(xdrs, &objp->obj_state.orphans) ||
           !xdr_int(xdrs, &objp->obj_state.datatype) ||
           !xdr_int(xdrs, &objp->obj_state.dleng) ||
           !xdr_int(xdrs, &objp->obj_state.datatyp2) ||
           !xdr_int(xdrs, &objp->obj_state.dleng2) ||
           !xdr_int(xdrs, &objp->obj_class) 
          )
          return(FALSE);
          }
    if (outl) {
        if (outl < 140) return(FALSE);
        sprintf(outp, "\nOBJECT: %s  Class= %s  [Created: ",
            objp->obj_name,
            class_name[objp->obj_class]);
        print_time(endof(outp), objp->obj_state.maketime) ;
        strcat(outp, "]\n  ReadTime: ");
        print_time(endof(outp), objp->obj_state.currtime);
        strcat(outp, ", ClearTime: ");
        print_time(endof(outp), objp->obj_state.cleartime);
        print_delta(endof(outp), objp->obj_state.cleartime, 
                     objp->obj_state.currtime) ;
        sprintf(endof(outp), "\n  Total Count= %d (+%d orphans)\n",
            objp->obj_state.totalno - objp->obj_state.orphans, /*31Dec*/
            objp->obj_state.orphans) ;
        outp += (n = strlen(outp));
        outl -= n;
    }
    return(TRUE);
} /* ReadStd */

    /*
     *  FAO_list() --  I/O routine for single entry of freq-all or 
     *                      freq-only read data.
     *
     *     objstp: contains parameters of object
     *     mfbep:  pts to element that may be sent or received (determined by
     *             direction of xdrs), and perhaps (if outl > 0)
     *             displayed into buffer: char *outp.
     *
     *     class:  FAOBJ | FOOBJ | FOBOBJ
     *
     *      Note: assumes FODE is subset of FADE structure.
     */
boolean FAO_list(xdrs, outp, outl, faep, objstp, enumh, class)
    register XDR  *xdrs;
    char *outp;
    int outl;
    FADE  *faep;
    struct Obj_state *objstp;
    int   enumh;
    int class;
    {
    register char *cp = outp;
    char *Enump;
    int i;
    char bytbuf[20];
    FOBDE *fobep = (FOBDE *) faep;
    char *valp = (class == FOBOBJ)? (char *) fobep->fob_value: 
                                    (char *) faep->fad_value;
    
    if (faep == NULL) return(FALSE);
    if (xdrs) {
        if (class == FOBOBJ) {
            if (!xdr_u_long(xdrs, &fobep->fob_count) ||
                !xdr_opaque(xdrs, &fobep->fob_Hbyte, sizeof(struct dblword)) ||
                !xdr_opaque(xdrs, fobep->fob_value, objstp->dleng) )
                     return(FALSE);
        }
        else { 
            if (!xdr_u_long(xdrs, &faep->fad_count) ||
                !xdr_opaque(xdrs, faep->fad_value, objstp->dleng))
                    return(FALSE);
            if (class == FAOBJ)  {
                if (!xdr_long(xdrs, &faep->fad_time))
                    return(FALSE);
            }
        }
    }
        
    if (outl) {
        if (outl < 80) return(FALSE);
        *cp++ = '[';
        print_val(cp, valp, objstp->datatype, objstp->dleng);
        if (Enump = (char *) FindEnum(enumh, valp, objstp->dleng)) {
            strcat(cp, " \"");
            strcat(cp,  Enump);
            strcat(cp, "\"");
            cp = endof(cp);
        }
        sprintf(endof(cp), "]= %d ", faep->fad_count); /* (same for either) */
        if (class == FOBOBJ) {
            print_dbl(bytbuf, fobep->fob_Hbyte, fobep->fob_Lbyte);
            sprintf(endof(cp), "&%sB ", bytbuf);
        }
       if ((i = Align(outp)) != 0)
           sprintf(endof(cp), "%*c", i, ' ');
        print_percent(endof(cp), objstp->totalno, faep->fad_count);
        if (class == FAOBJ)   /* FA -- add "@- kkksecs" */
           sprintf(endof(cp), "@-%dsec ", objstp->currtime - faep->fad_time);
        
        MakeCols(outp);
    }
    return(TRUE);
} /* FAO_list */


    /*
     *  MA_list() --  I/O routine for single entry of matrix-all read data
     *
     */
boolean MA_list(xdrs, outp, outl, maep, objstp, enumh)
    register XDR  *xdrs;
    char *outp;
    int outl;
    MADE  *maep;
    struct Obj_state *objstp;
    int    enumh;
{
    register char *cp = outp;
    char *Enump;
    int i;
    
    if (maep == NULL) return(FALSE);
    if (xdrs) {
        if (!xdr_u_long(xdrs, &maep->mad_count) ||
            !xdr_opaque(xdrs, maep->mad_value, objstp->dleng+objstp->dleng2) || 
            !xdr_long(xdrs, &maep->mad_time) )
            return(FALSE);
    }
        
    if (outl) {
        if (outl < 80) return(FALSE);
        *cp++ = '[';
        print_val(cp, maep->mad_value, objstp->datatype, objstp->dleng);
        if (Enump = (char *) FindEnum(enumh, maep->mad_value, objstp->dleng)) {
            strcat(cp, " \"");
            strcat(cp,  Enump);
            strcat(cp, "\"");
            cp = endof(cp);
        }
        strcat(cp, (tersemode)?":":" : ");
        print_val(endof(cp), maep->mad_value+objstp->dleng, objstp->datatyp2, 
                                 objstp->dleng2);
        
        sprintf(cp = endof(cp), "]= %d ", maep->mad_count);
        if ((i = Align(outp)) != 0)
           sprintf(endof(cp), "%*c", i, ' ');
           
        print_percent(endof(cp), objstp->totalno, maep->mad_count);
        sprintf(endof(cp), "@-%dsec ", objstp->currtime - maep->mad_time);
        MakeCols(outp);
    }
    return(TRUE);
} /* MA_list */


#define OUTSIZE 120
    /*
     *  ReadFO() --  Main I/O routine for reading FO (freq-only) and 
     *               FOB (freq-only-bytes) objects.
     *
     */
boolean ReadFO(xdrs, binpp, maxnop, defaultp, objp, binsize, islog, totbytp)
    register XDR *xdrs;
    FODE **binpp;
    int   *maxnop, *defaultp;
    OBJ   *objp;
    int    binsize, islog;
    struct FOparmB  *totbytp;  /* (NULL ptr if FO object) */
{
    register int i; 
    int   k = *maxnop;
    FODE *foep, fode;
    char  outp[OUTSIZE];
    int   outl = 0;
    int   enumh = ExistEnumObj(objp->obj_name);
    char  bytbuf[18];

    if (islog) {
        outl = OUTSIZE;
        MakeCols(NULL);
    }
    if (xdrs) {
       if (!xdr_u_long(xdrs, defaultp))
           return(FALSE);
       if (totbytp) {
          if (!xdr_opaque(xdrs, totbytp, sizeof(struct FOparmB)))
              return(FALSE);
       }
       if (!xdr_int(xdrs, &k))
          return(FALSE);     
       }
    if (islog) {    
        if (totbytp) {
            print_dbl(bytbuf, totbytp->totalB.H, totbytp->totalB.L);
            sprintf(outp, "  Total Bytes= %sB  #bins= %d\n", bytbuf, k);
        }
        else 
            sprintf(outp, "  #bins= %d\n", k);
        log(outp, logfp);
        sprintf(outp, "[default]= %d ", *defaultp);
        if (totbytp) {
            print_dbl(bytbuf, totbytp->defaultB.H, totbytp->defaultB.L);
            sprintf(endof(outp), "&%sB ", bytbuf);
        }
        if ((i = Align(outp)) != 0)
           sprintf(endof(outp), "%*c", i, ' ');
        print_percent(endof(outp), objp->obj_state.totalno, *defaultp);
        strcat(outp, "\n");
        log(outp, logfp);
    }
    
    if (*binpp && k>*maxnop) return(FALSE);
        /* decode: caller gave us a buffer, but it was too small. */
    *maxnop = k;
    
    aligncnt = 0;
    foep = (binsize)? *binpp : &fode;
    for (i=0;i<k;i++) {
        if (!FAO_list(xdrs, outp, outl, foep, &(objp->obj_state), enumh, 
                                                      (totbytp)?FOBOBJ:FOOBJ))
            return(FALSE);
        if (islog)  log(outp, logfp);
        foep = (FODE *) ((char *) foep + binsize);
    }
    if (islog) log("\n", logfp);
    return(TRUE);
} /* ReadFO */


    /*
     *  MFB_list() --  I/O routine for single entry of freq-all-bytes or
     *                                                 matrix-all-bytes
     *     objstp: contains parameters of object
     *     mfbep:  pts to element that may be sent or received (determined by
     *             direction of xdrs), and perhaps (if outl > 0)
     *             displayed into buffer: char *outp.
     *
     */
boolean MFB_list(xdrs, outp, outl, mfbep, objstp, enumh)
    register XDR  *xdrs;
    char *outp;
    int outl;
    MFBDE  *mfbep;
    struct Obj_state *objstp;
    int    enumh;
    {
    register char *cp = outp;
    char *Enump;
    int i;
    char bytbuf[20];
    
    if (mfbep == NULL) return(FALSE);
    if (xdrs) {
        if (!xdr_u_long(xdrs, &mfbep->mfbd_count) ||
            !xdr_long(xdrs, &mfbep->mfbd_time) ||
            !xdr_opaque(xdrs, &mfbep->mfbd_Hbyte, 8) ||
            !xdr_opaque(xdrs, mfbep->mfbd_value, objstp->dleng+objstp->dleng2) 
           )
           return(FALSE);
    }
        
    if (outl) {
        if (outl < 80) return(FALSE);
        *cp++ = '[';
        print_val(cp, mfbep->mfbd_value, objstp->datatype, objstp->dleng);
        if (Enump = (char *) FindEnum(enumh, mfbep->mfbd_value, objstp->dleng)) {
            strcat(cp, " \"");
            strcat(cp,  Enump);
            strcat(cp, "\"");
            cp = endof(cp);
        }
        if (objstp->dleng2) {
            strcat(cp, (tersemode)?":":" : ");
            print_val(endof(cp), mfbep->mfbd_value+objstp->dleng, 
                                 objstp->datatyp2, objstp->dleng2);
        }
        print_dbl(bytbuf, mfbep->mfbd_Hbyte, mfbep->mfbd_Lbyte);
        sprintf(cp = endof(cp), "]= %d &%sB ", mfbep->mfbd_count,  bytbuf);
        if ((i = Align(outp)) != 0)
           sprintf(endof(cp), "%*c", i, ' ');
           
        print_percent(endof(cp), objstp->totalno, mfbep->mfbd_count);
        sprintf(endof(cp), "@-%dsec ", objstp->currtime - mfbep->mfbd_time);
        MakeCols(outp);
    }
    return(TRUE);
} /* MFB_list */


    /*
     *  ReadWS() --  I/O routine for working-set read data.
     *
     */
boolean ReadWS(xdrs, binpp, maxnop, distinctp, objp, islog)
    register XDR *xdrs;
    u_long  **binpp;
    int      *maxnop;
    u_long   *distinctp;
    OBJ      *objp;
    int      islog;
{
    register int i, n;
    int     k = *maxnop;
    char    outp[OUTSIZE];
    u_long  count, *wsep = (*binpp)? *binpp: &count;
            
    if (xdrs) {
        if (!xdr_u_long(xdrs, distinctp) ||  !xdr_int(xdrs, &k))
                  return(FALSE);
    }
    if (islog) {    
       sprintf(outp, " # distinct values= %d\n", *distinctp) ;
       log(outp, logfp);
    }
    
    if (k>*maxnop) return(FALSE);
        /* decode: caller gave us a buffer, but it was too small. */
    *maxnop = k;
        
    n = 1;  
    for (i=0;i < k; i++) {
        if (xdrs) {
            if (!xdr_u_long(xdrs, wsep))  return(FALSE);
        }
        if (islog) {
           sprintf(outp, 
              "[Cache misses - cache of %4d]= %6d   Prob.= %.2f\n", 
              n, *wsep, ((float) *wsep)/objp->obj_state.totalno);
            log(outp, logfp);
        }
        wsep++;
        n <<= 1;                        
    } ;
 
    return(TRUE);
} /* ReadWS */


    /*
     *  ReadP2() --  I/O routine for hist_pwr2 read data.
     *
     */
boolean ReadP2(xdrs, binpp, maxnop, p2dp,  islog)
    register XDR *xdrs;
    HistE  **binpp;
    int     *maxnop;
    P2DATA  *p2dp;
    int      islog;
{
    register u_long s;
    int k = *maxnop;
    char outp[OUTSIZE];
    int incr = (*binpp)? sizeof(u_long): 0;
    HistE count, *hep = (*binpp)? *binpp: &count;
        
    if (!readhist(xdrs, p2dp, islog)) return(FALSE);
    
    if (xdrs) {
       if (!xdr_int(xdrs, &k) ||
           !xdr_u_long(xdrs, hep)
           )
          return(FALSE);
       }
    if (islog) {    
       s = p2dp->Hist_scalef;
       sprintf(outp, "  [0-%d]= %d\n",  s-1, *hep);
       log(outp, logfp);
    }
    
    if (*binpp&&k>*maxnop) return(FALSE);
        /* decode: caller gave us a buffer, but it was too small. */
    *maxnop = k;
    
    while (--k > 0) {
        hep = (u_long *) ((char *) hep + incr); 
        if (xdrs) 
           if (!xdr_u_long(xdrs, hep))  return(FALSE);
        if (islog) {
            if (k)
               sprintf(outp, "  [%d-%d]= %d\n", s, s+s-1, *hep);
            else
               sprintf(outp, "  [%d-inf'ty]= %d\n", s, *hep) ;
            s <<= 1;
            log(outp, logfp);
        };
    }
 
    return(TRUE);
} /* ReadP2 */


    /*
     *  ReadHI() --  I/O routine for 'hist' class read data.
     */
boolean ReadHI(xdrs, binpp, maxnop, Histp,  islog)
    register XDR *xdrs;
    HistE  **binpp;
    int     *maxnop;
    P2DATA  *Histp;
    int      islog;
{
    register u_long s = 0;
    int k = *maxnop;
    char outp[OUTSIZE];
    int incr = (*binpp)? sizeof(HistE): 0;
    HistE count, *hep = (*binpp)? *binpp: &count;
    
    if (!readhist(xdrs, Histp, islog)) return(FALSE);
    
    if (xdrs) {
       if (!xdr_int(xdrs, &k))  /* bin count */
          return(FALSE);
       }
    
    if (*binpp&&k>*maxnop) return(FALSE);
        /* decode: caller gave us a buffer, but it was too small. */
    *maxnop = k;
    
    while (k-- > 0) {
        if (xdrs) 
           if (!xdr_u_long(xdrs, hep))  return(FALSE);
        if (islog) {
            sprintf(outp, "  [%d-%d]= %d\n", 
                              s, s+Histp->Hist_scalef-1, *hep);
            s += Histp->Hist_scalef;
            log(outp, logfp);
        };
        hep = (HistE *) ((char *) hep + incr);  
    }
    if (islog && Histp->Hist_default) { 
        sprintf(outp, "  [Off-scale]= %d\n", Histp->Hist_default);
        log(outp, logfp);
    }
 
    return(TRUE);
} /* ReadHI */


    /* readhist() -- Common histogram routine used by ReadHI(), ReadP2().
     *
     */
boolean readhist(xdrs, hdp, islog)
    register XDR *xdrs;
    HistDATA *hdp;
    int      islog;
{
    char outp[OUTSIZE];
        
    if (xdrs) {
       if (!xdr_u_long(xdrs, &hdp->Hist_scalef) ||
           !xdr_u_long(xdrs, &hdp->Hist_avg) ||
           !xdr_u_long(xdrs, &hdp->Hist_min) ||
           !xdr_u_long(xdrs, &hdp->Hist_max) ||
           !xdr_u_long(xdrs, &hdp->Hist_default) 
           )
        return(FALSE);
    }
    if (islog) 
        if (hdp->Hist_min != INFINITY) {       
            sprintf(outp,  "   Avg= %d  Min= %d  Max= %d\n", 
               hdp->Hist_avg, hdp->Hist_min, hdp->Hist_max);
            log(outp, logfp);
        }
    return(TRUE);
} /* readhist */    
           
           
    /*
     *  ReadBP() -- I/O routine for bin-pkt class read
     *
     */
boolean ReadBP(xdrs, bpep, objp, islog)
    XDR  *xdrs;
    register BPE *bpep;
    OBJ  *objp;
    int   islog;
{
    char *cp = bpep->bpe_packet;
    int   len = bpep->bpe_pktlen;
    char  out[256];
    long deltat;

    if (xdrs) {
       if (!xdr_int(xdrs, &bpep->bpe_time) ||
           !xdr_bytes(xdrs, &cp, &len, MAX_BPSIZE)
          )
          return(FALSE);
       }    
    bpep->bpe_pktlen = len;
    if (islog) {
        deltat =  objp->obj_state.currtime - bpep->bpe_time;
        sprintf(out, " @-%dsec >> ", deltat);       
        if (objp->obj_state.datatype == DTYP_packet) {
            strcat(out, "\n");
            log(out, logfp);
            hex_print(cp, len);
        }
        else {
            print_val(endof(out), cp, objp->obj_state.datatype, len);
            strcat(out, "\n");
            log(out, logfp);
        }
    }
    return(TRUE);
} /* ReadBP */

    
    
/*       
 *       boolean xdr_invoke() --
 *
 *   Routine to do XDR encode/decode for invocation request
 *             within ATTACH command.
 */
boolean
xdr_invoke(xdrs, invp, valp)
    XDR *xdrs;
    struct invokes *invp;
    u_long *valp;
    {
    char *cpf1 = invp->inv_field1,
         *cpf2 = invp->inv_field2,
         *cpobj= invp->inv_objname;
    int parmbytes = 4*(invp->inv_nparms);

    if( !xdr_string(xdrs, &cpf1, MAX_FLDNAME) ||
        !xdr_string(xdrs, &cpf2, MAX_FLDNAME) ||
        !xdr_string(xdrs, &cpobj, MAX_OBJNAME) ||
        !xdr_int(xdrs, &invp->inv_classno))
        return(FALSE);
    if (valp) {
        if (!xdr_bytes(xdrs, &valp, &parmbytes,
                                     MAXPARMS*LONGPERVAL*sizeof(u_long))) 
            return(FALSE);
        invp->inv_nparms = parmbytes/4;  /*19Feb88*/
    }
    return(TRUE);
} /* xdr_invoke() */


    /*--------------------------------------------------------------
     *  
     *               I/O Routines for SHOW command
     *
     *   Routines to do XDR encode/decode and/or display for SHOW command.
     *
     *          SHOWfield()
     *          SHOWsublist()
     *          SHOWinv()
     */    
/*
 *  SHOWfield() -- Outer loop I/O routine for SHOW command. 
 *      Called for each new AT field
 *          char *fname -- Name of field at which invocation will occur
 *          int   fno   -- Corresponding field number
 */
boolean
SHOWfield(xdrs, fname, fno, islog)
    XDR   *xdrs;
    char  *fname;
    int    fno, islog;
    {
    char  out[256];
    
    if (xdrs)
        if (!xdr_string(xdrs, &fname, MAX_FLDNAME)||
            !xdr_int(xdrs, &fno))
                return(FALSE);
    if (islog) {
        sprintf(out, "INVOKED AT %.32s (FIELD #%d):\n", fname, fno);
        log(out, logfp);
    }
    
    return(TRUE);
}  /* SHOWfield() */

/*
 *  SHOWsublist() -- TRUE/FALSE Sublist I/O routine for SHOW command .
 *      Called at start of each IF sublist
 *         boolean parity -- select sublist TRUE/FALSE.
 *         int level      -- integer level in tree
 */
boolean
SHOWsublist(xdrs, parity, level, islog) 
    XDR  *xdrs;
    boolean parity;
    int   level, islog;
    {
    char  out[256];
    int   code = (parity)? StartTsubl: StartFsubl;

    if (xdrs)       
        if (!xdr_int(xdrs, &code))  
            return(FALSE);
        
    if (islog) {
        sprintf(out, "%*c %s=>\n", 3*level-2, ' ',
            (parity)? "True":"False" );
        log(out, logfp);
    }
    return(TRUE);
} /* SHOWsublist() */
        
    
/*
 *  SHOWinv() -- Invocation I/O routine for SHOW command.
 *      Called for each invocation
 *          struct invokes *invokesp -- ptr to invocation parm structure
 *          int level --  invocation level number
 */
boolean
SHOWinv(xdrs, invokesp, level, islog)
    XDR *xdrs;
    register struct invokes *invokesp;
    int   level, islog;
    {
    char  out[256];
    int indent = 3*level+2;
    extern boolean is_filter[];
                        
    if (xdrs) 
        if (!xdr_invoke(xdrs, invokesp, NULL))
                return(FALSE);
    
    if (islog) {
        sprintf(out, "%*c", indent, ' ');
        if (is_filter[invokesp->inv_classno])
            sprintf(&out[indent], "IF %.32s IS (FILTER_OBJECT)",
                invokesp->inv_field1);
        else        
            sprintf(&out[indent], "RECORD %.32s %.32s IN (OBJECT)",
                invokesp->inv_field1, invokesp->inv_field2);

        sprintf(endof(out), " %.32s (CLASS) %s\n",
            invokesp->inv_objname, class_name[invokesp->inv_classno]);      
        log(out, logfp);
    }
    return(TRUE);
} /* SHOWinv() */



/********************************************************************
 *
 *                 UTILITY ROUTINES
 *
 ********************************************************************/
 
static char *IPformat[] = {"%d.%d.%d.%d", "%d.%d.%d", "%d.%d", "%d"};
                     
    /*
     *  Print value, given ptr to first byte and length and data type.
     *    Note: bytes are in big-endian (network) order.
     */
void     
print_val(val_str, valuep, dtype, dleng)
   char *val_str;
   register u_char *valuep;
   int   dtype;
   register int dleng;
{
   u_long v_wd;
   register int i;
   register u_char *cp;
    
   switch (dtype) {
         
      case DTYP_IP:  /* IP address ... assume length == 4 */
         if (tersemode) 
             {  /* Terse ... set i # of trailing zero bytes */
             cp = valuep+3;
             for (i=0; i<3; i++) if (*cp--) break;
         }
         else i = 0;
         sprintf(val_str, IPformat[i], 
                 0xff&*valuep, 0xff&*(valuep+1), 
                 0xff&*(valuep+2), 0xff&*(valuep+3));
         break;
              
      case DTYP_EtherA:  /* Ethernet Address ...   */
         strcpy(val_str, ether_ntoa(valuep)) ;
         break;
         
      case DTYP_int:  /* Integer... length <= 4 */
         if (dleng <= 4) {
             v_wd = 0;
             bcopy(valuep, &v_wd, dleng);  /* align on fullword boundary... */
             sprintf(val_str, "%d", 
                    (dleng==1)?  * (u_char *) &v_wd :
                    (dleng==2)? ntohs(*(u_short *) &v_wd) : 
                                ntohl(v_wd) );
             break;
         }  /* else print in hex... */
         
      case DTYP_bits:
      case DTYP_unknown: 
      default:  /* Print in hex */
         strcpy(val_str--, "x");
         while (dleng-- > 0)   
            sprintf(val_str += 2, "%2.2x", *valuep++);
         break;         
    }  /* end switch */
} /* print_val */
 
 /*
  *  print_dbl(buffer, High-order4, Low-order3);
  *
  */
void
print_dbl(p, H, L)
    char *p;
    register u_long H, L;
    {
    register char *cp = p;
    u_long N;
    
    if (L == 0 && H == 0) {
        *p++ = '0';
        *p = '\0';
        return;
    }
    p += 18;   
    *p = '\0';    
    while ((H)||(L)){
        N = ((H%10)<<24)+L; 
        H = H/10;
        L = N/10;
        *--p = "0123456789"[N%10];
    }
    while (*cp++ = *p++);        
       
}  /* print_dbl() */

    
void
hex_print(p, len)
   register char *p;
   register int len;
{
   register int i = 0, j ;
   char out[128];
         
   while (len > 0) {
      for (j = 0; (j<4) && (len > 0); j++)  {
         if (!j) 
           sprintf(out, "x%4.4x: ", i );
           strcat(out, " ");

           print_val(endof(out), p, DTYP_bits, 
                 (len < sizeof(u_long))?len:sizeof(u_long));
           p += sizeof(u_long) ;
           i += sizeof(u_long) ;
           len -= sizeof(u_long) ;
      }
      strcat(out, "\n");
      log(out, logfp);
   }
   
} /* hex_print */

void
hex_dump(xdrs, tracefp, outp, outl, max, disp_type)
   XDR *xdrs;
   FILE *tracefp;
   char *outp;
   int outl,          /* number of characters */
       max,           /* number of words */
       disp_type;
{
   register int i = 0, j ;
   u_long temp ;
   boolean exit = FALSE; 
         
   while (! exit) {
      for (j = 0; j<4; j++)  {
         if (max && (((i*4)+j) >= max)) {
            exit = TRUE;
            break;
            }
         if (xdr_u_long(xdrs, &temp)) {
            if (!j) 
               sprintf(outp,(disp_type == DTYP_int)? "%3d: ":
                           (disp_type == DTYP_IP)?  "%3d: ":
                                                    "x%2.2x: ", i );
            strcat(outp, " ");
            print_val(endof(outp), &temp, disp_type, sizeof(u_long));
            }
         else {
            exit = TRUE;
            break;
            }
         }
      if (tracefp) fprintf(tracefp,"%s\n", outp);
      *outp = '\0';
      i++;
      }
} /* hex_dump */

void
print_time(out_str, i)
   char *out_str;
   long i;
{
/*  
     struct tm *localtime(clock)
          long *clock;
     char *asctime(tm)
          struct tm *tm;
    
     struct tm {
        int tm_sec;           -- seconds (0 - 59) --
        int tm_min;           -- minutes (0 - 59) --
        int tm_hour;          -- hours (0 - 23) --
        int tm_mday;          -- day of month (1 - 31) --
        int tm_mon;           -- month of year (0 - 11) --
        int tm_year;          -- year - 1900 --
        int tm_wday;          -- day of week (Sunday = 0) --
        int tm_yday;          -- day of year (0 - 365) --
        int tm_isdst;
            };

    struct timezone {
        int  tz_minuteswest;     -- of Greenwich --
        int  tz_dsttime;    -- type of dst correction to apply --
        };
        
*/  
   struct tm *t;

   if (print_utime)
      t = gmtime(&i);
   else
      t = localtime(&i);
   sprintf(out_str, "%2.2d:%2.2d:%2.2d %2.2d-%2.2d-%2.2d", 
      t->tm_hour, t->tm_min, t->tm_sec,
      t->tm_mon+1, t->tm_mday, t->tm_year);
      
} /* print_time */   

void
print_delta(out_str, time1, time2)
   char *out_str;
   long time1, time2;
{
   sprintf(out_str, " (@-%dsec)", time2 - time1);
} /* print_delta */
    
    /*
     *  Given ptr to null-terminated string containing new item in 
     *  output stream, append enough blanks to make output
     *  line up in nice columns.  If there is not room on current line
     *  for another max-sized line, output NL, instead.  Global
     *  variables MaxItemSize and LineUsed are needed.
     */
void     
MakeCols(cp)
    register char *cp;
{
    register int W, i, MIS;
    char temp[120];
    
    if (cp == NULL) {
        /* Initialize */
        MaxItemSize = LineUsed = aligncnt = prevW = 0;
        return;
    }
    
    W = strlen(cp);
    if (tersemode) LineUsed = MAXCOLS+1;  /* Terse mode: force newline always */
    else if (!LineUsed) {
       /* line is empty */
       prevW = LineUsed = W;
       return;
       }

    MIS = ( ((prevW > MaxItemSize)? prevW: MaxItemSize)+7)&~7;
    strcpy(temp, cp);

    i = MIS - prevW;
    LineUsed += i+W;
    if (LineUsed > MAXCOLS) {
       /* it wouldn't fit */
       sprintf(cp,"\n%s",temp);
       prevW = LineUsed = W;
       }
    else {
       /* another column fit */
       MaxItemSize = MIS-7;
       if (i > 0) {
          sprintf(cp, "%*c", i, ' ');
          strcat(cp, temp);
          }
       prevW = W;
       }
       
} /* MakeCols */

int Align(cp)
    char *cp;
    {
    register int n = strlen(cp)+1;
    register int k;
    
    if (tersemode) return(0);
    aligncnt = (aligncnt) ? (aligncnt + (n - (aligncnt>>3))) :
        /* That is 8*(exponential avg of n), decay factor 1/8 */
                       (n<<3);
        /* Initial value  = 8*n */
        
    k = ((((n > (k = (aligncnt>>3)))? n : k)+3)&~3);
    return(k-n);
} /* Align */

void
print_percent(cp, total, part)
   char *cp;
   int total, part;
{
   float rate;

   if (tersemode) {  /* In terse mode, suppress percentages */
       *cp = '\0';
       return;
   }
   rate = (total)? ((((float) part)/total)*100): 0;

   if (rate == 0)
      sprintf(cp, "(0%%%%) ");   
   else if (rate < .1)
      sprintf(cp, "( <.1%%%%) ");
   else if (rate < 10)
      sprintf(cp, "( %.1f%%%%) ", rate); 
   else
      sprintf(cp, "(%.1f%%%%) ", rate);
} /* print_percent */
